"""
This test file tests the lib.policy.py

The lib.policy.py only depends on the database model.
"""

PW_FILE_2 = "tests/testdata/passwords"
DICT_FILE = "tests/testdata/dictionary"

import datetime
from datetime import timedelta

from flask import g

from privacyidea.lib.authcache import delete_from_cache, _hash_password
from privacyidea.lib.error import UserError, PolicyError
from privacyidea.lib.policy import (set_policy, delete_policy,
                                    PolicyClass, SCOPE,
                                    ACTION, ACTIONVALUE, LOGINMODE)
from privacyidea.lib.policydecorators import (auth_otppin,
                                              auth_user_does_not_exist,
                                              auth_user_passthru,
                                              auth_user_has_no_token,
                                              login_mode, config_lost_token,
                                              auth_cache,
                                              auth_lastauth, reset_all_user_tokens)
from privacyidea.lib.radiusserver import add_radius
from privacyidea.lib.realm import set_realm, delete_realm
from privacyidea.lib.resolver import save_resolver, delete_resolver
from privacyidea.lib.token import (init_token, remove_token, check_user_pass,
                                   get_tokens)
from privacyidea.lib.user import User
from privacyidea.models import AuthCache
from . import radiusmock
from .base import MyTestCase, FakeFlaskG, FakeAudit


def _check_policy_name(policy_name, policies):
    """
    Checks if the policy_name is contained in the policies list.
    """
    contained = False
    for policy in policies:
        if policy.get("name") == policy_name:
            contained = True
            break
    return contained


class LibPolicyTestCase(MyTestCase):
    """
    Test all the internal libpolicy decorators
    """

    @staticmethod
    def fake_check_otp(dummy, pin, user=None, options=None):
        return pin == "FAKE"

    @staticmethod
    def fake_check_token_list(tokenobject_list, passw, user=None, options=None, allow_reset_all_tokens=True,
                              result=False):
        return result, "some text"

    def test_01_otppin(self):
        my_user = User("cornelius", realm="r1")
        set_policy(name="pol1", scope=SCOPE.AUTH, action=f"{ACTION.OTPPIN}={ACTIONVALUE.NONE}")
        fake_g = FakeFlaskG()
        policy_class = PolicyClass()
        fake_g.policy_object = policy_class
        fake_g.audit_object = FakeAudit()
        options = {"g": fake_g}

        # NONE with empty PIN -> success
        r = auth_otppin(self.fake_check_otp, None, "", options=options, user=my_user)
        self.assertTrue(r)
        # NONE with empty PIN -> success, even if the authentication is done
        # for a serial and not a user, since the policy holds for all realms
        token = init_token({"type": "HOTP", "otpkey": "1234"})
        r = auth_otppin(self.fake_check_otp, token, "", options=options, user=None)
        self.assertTrue(r)

        # NONE with some pin -> fail
        r = auth_otppin(self.fake_check_otp, None, "some pin", options=options, user=my_user)
        self.assertFalse(r)

        delete_policy("pol1")
        set_policy(name="pol1", scope=SCOPE.AUTH, action=f"{ACTION.OTPPIN}={ACTIONVALUE.TOKENPIN}")
        fake_g = FakeFlaskG()
        policy_class = PolicyClass()
        fake_g.policy_object = policy_class
        fake_g.audit_object = FakeAudit()
        options = {"g": fake_g}

        r = auth_otppin(self.fake_check_otp, None, "FAKE", options=options, user=my_user)
        self.assertTrue(r)
        r = auth_otppin(self.fake_check_otp, None, "Wrong Pin", options=options, user=my_user)
        self.assertFalse(r)
        delete_policy("pol1")

    def test_02_userstore_password(self):
        # create a realm, where cornelius has a password test
        rid = save_resolver({"resolver": "myreso",
                             "type": "passwdresolver",
                             "fileName": PW_FILE_2})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm("r1", [{'name': "myreso"}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        # now create a policy with userstore PW
        set_policy(name="pol1", scope=SCOPE.AUTH, action=f"{ACTION.OTPPIN}={ACTIONVALUE.USERSTORE}")
        fake_g = FakeFlaskG()
        policy_class = PolicyClass()
        fake_g.policy_object = policy_class
        fake_g.audit_object = FakeAudit()
        options = {"g": fake_g}
        # Wrong password
        r = auth_otppin(self.fake_check_otp, None,
                        "WrongPW", options=options,
                        user=User("cornelius", realm="r1"))
        self.assertFalse(r)

        # Correct password from userstore: "test"
        r = auth_otppin(self.fake_check_otp, None,
                        "test", options=options,
                        user=User("cornelius", realm="r1"))
        self.assertTrue(r)
        delete_policy("pol1")

    def test_03_otppin_for_serial(self):
        # now create a policy with userstore PW
        set_policy(name="pol1", scope=SCOPE.AUTH, action=f"{ACTION.OTPPIN}={ACTIONVALUE.USERSTORE}")
        fake_g = FakeFlaskG()
        policy_class = PolicyClass()
        fake_g.policy_object = policy_class
        fake_g.audit_object = FakeAudit()
        options = {"g": fake_g, "serial": "T001"}

        # Create a token and assign to user cornelius
        token = init_token({"serial": "T001", "type": "hotp", "genkey": 1},
                           user=User("cornelius", realm="r1"))
        self.assertTrue(token)

        # Wrong password
        # Not identified by the user but by the token owner
        r = auth_otppin(self.fake_check_otp, token, "WrongPW", options=options, user=None)
        self.assertFalse(r)

        # Correct password from userstore: "test"
        # Not identified by the user but by the token owner
        r = auth_otppin(self.fake_check_otp, token,
                        "test", options=options,
                        user=None)
        self.assertTrue(r)
        delete_policy("pol1")
        remove_token("T001")

    def test_04_user_does_not_exist(self):
        user = User("MisterX", realm="NoRealm")
        passw = "wrongPW"
        options = {}
        # A non-existing user will fail to authenticate without a policy
        self.assertRaises(UserError, auth_user_does_not_exist,
                          check_user_pass, user, passw, options)

        # Now we set a policy, that a non existing user will authenticate
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   action=ACTION.PASSNOUSER)
        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}
        rv = auth_user_does_not_exist(check_user_pass, user, passw,
                                      options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "user does not exist, accepted due "
                         "to 'pol1'")
        delete_policy("pol1")

    def test_04a_user_does_not_exist_without_resolver(self):
        user = User("MisterX", realm=self.realm1)
        passw = "somePW"

        # Now we set a policy, that a non existing user will authenticate
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   action="{0}, {1}, {2}, {3}=none".format(
                       ACTION.RESETALLTOKENS,
                       ACTION.PASSNOUSER,
                       ACTION.PASSNOTOKEN,
                       ACTION.OTPPIN
                   ),
                   realm=self.realm1)
        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}
        rv = auth_user_does_not_exist(check_user_pass, user, passw,
                                      options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "user does not exist, accepted due "
                         "to 'pol1'")
        delete_policy("pol1")

    def test_05_user_has_no_tokens(self):
        user = User("cornelius", realm="r1")
        passw = "test"
        options = {}
        # A user with no tokens will fail to authenticate
        rv = auth_user_has_no_token(check_user_pass, user, passw, options)
        self.assertFalse(rv[0])
        self.assertEqual(rv[1].get("message"), "The user has no tokens assigned")

        # Now we set a policy, that a non-existing user will authenticate
        set_policy(name="pol1", scope=SCOPE.AUTH, action=ACTION.PASSNOTOKEN)
        fake_g = FakeFlaskG()
        fake_g.policy_object = PolicyClass()
        fake_g.audit_object = FakeAudit()
        options = {"g": fake_g}
        rv = auth_user_has_no_token(check_user_pass, user, passw, options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"), "user has no token, accepted due to 'pol1'")
        delete_policy("pol1")

    @radiusmock.activate
    def test_06_passthru(self):
        user = User("cornelius", realm="r1")
        passw = "test"
        options = {}
        # A user with no tokens will fail to authenticate
        self.assertEqual(get_tokens(user=user, count=True), 0)
        rv = auth_user_passthru(check_user_pass, user, passw, options)
        self.assertFalse(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "The user has no tokens assigned")

        # Now we set a PASSTHRU policy, so that the user may authenticate
        # against his userstore (old style)
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   action=ACTION.PASSTHRU)
        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}
        rv = auth_user_passthru(check_user_pass, user, passw,
                                options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "against userstore due to 'pol1'")

        # Now set a PASSTHRU policy to the userstore (new style)
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   action="{0!s}=userstore".format(ACTION.PASSTHRU))
        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}
        rv = auth_user_passthru(check_user_pass, user, passw, options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"), "against userstore due to 'pol1'")

        # Now set a PASSTHRU policy to a RADIUS config (new style)
        radiusmock.setdata(response=radiusmock.AccessAccept)
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   action="{0!s}=radiusconfig1".format(ACTION.PASSTHRU))
        r = add_radius("radiusconfig1", "1.2.3.4", "testing123", dictionary=DICT_FILE)
        self.assertTrue(r > 0)

        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}
        rv = auth_user_passthru(check_user_pass, user, passw, options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "against RADIUS server radiusconfig1 due to 'pol1'")

        # Now assign a token to the user. If the user has a token and the
        # passthru policy is set, the user must not be able to authenticate
        # with his userstore password.
        init_token({"serial": "PTHRU",
                    "type": "spass", "pin": "Hallo"},
                   user=user)
        rv = auth_user_passthru(check_user_pass, user, passw,
                                options=options)
        self.assertFalse(rv[0])
        self.assertEqual(rv[1].get("message"), "wrong otp pin")

        remove_token("PTHRU")
        delete_policy("pol1")

    def test_07_login_mode(self):
        # a realm: cornelius@r1: PW: test

        def check_webui_user_userstore(user_obj, password,
                                       options=None, superuser_realms=None,
                                       check_otp=False):
            self.assertEqual(check_otp, False)

        def check_webui_user_privacyidea(user_obj, password,
                                         options=None, superuser_realms=None,
                                         check_otp=False):
            self.assertEqual(check_otp, True)

        user_obj = User("cornelius", "r1")

        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        # No policy, the function is called with check_otp=False
        login_mode(check_webui_user_userstore, user_obj, "",
                   options=options, superuser_realms="", check_otp=False)

        set_policy(name="pol2",
                   scope=SCOPE.WEBUI,
                   action="{0!s}={1!s}".format(ACTION.LOGINMODE, LOGINMODE.PRIVACYIDEA))
        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        # Policy is set, the function is called with check_otp=True
        login_mode(check_webui_user_privacyidea, user_obj, "",
                   options=options, superuser_realms="", check_otp=False)

        # Set policy, so that the user is not allowed to login at all
        set_policy(name="pol2",
                   scope=SCOPE.WEBUI,
                   action="{0!s}={1!s}".format(ACTION.LOGINMODE, LOGINMODE.DISABLE))
        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        # Policy is set. Trying to login raises a policy error
        self.assertRaises(PolicyError, login_mode,
                          check_webui_user_privacyidea, user_obj, "",
                          options=options, superuser_realms="",
                          check_otp=False)
        delete_policy("pol2")

    def test_08_config_lost_token_policy(self):
        def func1(serial, validity=10, contents="Ccns", pw_len=16,
                  options=None):
            self.assertEqual(validity, 10)
            self.assertEqual(contents, "Ccns")
            self.assertEqual(pw_len, 16)

        def func2(serial, validity=10, contents="Ccns", pw_len=16,
                  options=None):
            self.assertEqual(validity, 5)
            self.assertEqual(contents, "C")
            self.assertEqual(pw_len, 3)

        init_token({"serial": "LOST001", "type": "hotp", "genkey": 1},
                   user=User("cornelius", realm="r1"))

        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        # No policy, the function is called with default values
        config_lost_token(func1, "LOST001", options=options)

        set_policy(name="lost_pol2",
                   scope=SCOPE.ENROLL,
                   action="%s=%s, %s=%s,"
                          "%s=%s" % (ACTION.LOSTTOKENPWCONTENTS, "C",
                                     ACTION.LOSTTOKENVALID, 5,
                                     ACTION.LOSTTOKENPWLEN, 3))
        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        # Policy is set, the function is called with check_otp=True
        config_lost_token(func2, "LOST001", options=options)

    def test_09_challenge_response_allowed(self):
        user = User("cornelius", realm="r1")
        pin = "test"

        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        options = {"g": g}
        token = init_token({"type": "hotp",
                            "otpkey": "1234",
                            "pin": pin}, user=user)
        # With no policy, it will be no chal resp
        rv = token.is_challenge_request(pin, user=user, options=options)
        self.assertEqual(rv, False)

        # Now we set a policy with several tokentypes
        set_policy(name="pol_chal_resp_1",
                   scope=SCOPE.AUTH,
                   action="{0!s}=hotp tiqr totp".format(ACTION.CHALLENGERESPONSE))
        set_policy(name="pol_chal_resp_2",
                   scope=SCOPE.AUTH,
                   action="{0!s}=hotp motp".format(ACTION.CHALLENGERESPONSE))
        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}
        rv = token.is_challenge_request(pin, user=user, options=options)
        self.assertEqual(rv, True)
        delete_policy("pol_chal_resp_1")

    def test_10_auth_lastauth(self):
        serial = "SPASSLASTAUTH"
        pin = "secretpin"

        def fake_auth_missing_serial(user, pin, options=None):
            return True, {}

        def fake_auth(user, pin, options):
            return True, {"serial": serial}

        user = User("cornelius", realm="r1")
        init_token({"type": "spass",
                    "pin": pin,
                    "serial": serial}, user=user)

        # set time limit policy
        set_policy(name="pol_lastauth",
                   scope=SCOPE.AUTHZ,
                   action="{0!s}=1d".format(ACTION.LASTAUTH))
        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}

        rv = auth_lastauth(fake_auth, user, pin, options)
        self.assertEqual(rv[0], True)

        token = get_tokens(serial=serial)[0]
        # Set a very old last_auth
        token.add_tokeninfo(ACTION.LASTAUTH,
                            datetime.datetime.utcnow() - datetime.timedelta(days=2))
        rv = auth_lastauth(fake_auth, user, pin, options)
        self.assertEqual(rv[0], False)
        self.assertTrue("The last successful authentication was" in
                        rv[1].get("message"), rv[1])

        remove_token(serial)
        delete_policy("pol_lastauth")

    def test_11_otppin_with_resolvers(self):
        # This tests, if the otppin policy differentiates between users in
        # the same realm but in different resolvers.
        r = save_resolver({"resolver": "reso001",
                           "type": "passwdresolver",
                           "fileName": "tests/testdata/passwords"})
        # user "cornelius" is in resolver reso001
        self.assertTrue(r > 0)
        r = save_resolver({"resolver": "reso002",
                           "type": "passwdresolver",
                           "fileName": "tests/testdata/pw-2nd-resolver"})
        # user "userresolver2" is in resolver reso002
        self.assertTrue(r > 0)
        (added, failed) = set_realm("myrealm",
                                    [
                                        {'name': "reso001"},
                                        {'name': "reso002"}])
        self.assertEqual(len(added), 2)
        self.assertEqual(len(failed), 0)
        my_user_1 = User("cornelius", realm="myrealm")
        my_user_2 = User("userresolver2", realm="myrealm")
        # We set a policy only for resolver reso002
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   realm="myrealm",
                   resolver="reso002",
                   action="{0!s}={1!s}".format(ACTION.OTPPIN, ACTIONVALUE.NONE))
        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        # user in reso001 fails with empty PIN, since the policy does not
        # match for him
        r = auth_otppin(self.fake_check_otp, None,
                        "", options=options, user=my_user_1)
        self.assertFalse(r)

        # user in reso002 succeeds with empty PIN, since policy pol1 matches
        # for him
        r = auth_otppin(self.fake_check_otp, None,
                        "", options=options, user=my_user_2)
        self.assertTrue(r)

        # user in reso002 fails with any PIN, since policy pol1 matches
        # for him
        r = auth_otppin(self.fake_check_otp, None,
                        "anyPIN", options=options, user=my_user_2)
        self.assertFalse(r)

        delete_policy("pol1")
        delete_realm("myrealm")
        delete_resolver("reso001")
        delete_resolver("reso002")

    def test_12_authcache(self):
        password = "secret123456"
        username = "cornelius"
        realm = "myrealm"
        resolver = "reso001"

        pwd_hash = _hash_password(password)

        r = save_resolver({"resolver": resolver,
                           "type": "passwdresolver",
                           "fileName": "tests/testdata/passwords"})
        set_realm(realm, [{'name': resolver}])

        def fake_check_user_pass(user, passw, options=None):
            return True, {"message": "Fake Authentication"}

        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   realm=realm,
                   resolver=resolver,
                   action="{0!s}={1!s}".format(ACTION.AUTH_CACHE, "4h/5m"))
        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        # This successfully authenticates against the authcache
        # We have an authentication, that is within the policy timeout
        AuthCache(username, realm, resolver, pwd_hash,
                  first_auth=datetime.datetime.utcnow() - timedelta(hours=3),
                  last_auth=datetime.datetime.utcnow() - timedelta(minutes=1)).save()
        r = auth_cache(fake_check_user_pass, User(username, realm),
                       password, options=options)
        self.assertTrue(r[0])
        self.assertEqual(r[1].get("message"), "Authenticated by AuthCache.")

        # We have an authentication, that is not read from the authcache,
        # since the authcache first_auth is too old.
        delete_from_cache(username, realm, resolver, password)
        AuthCache(username, realm, resolver, pwd_hash,
                  first_auth=datetime.datetime.utcnow() - timedelta(hours=5),
                  last_auth=datetime.datetime.utcnow() - timedelta(
                      minutes=1)).save()
        r = auth_cache(fake_check_user_pass, User(username, realm),
                       password, options=options)
        self.assertTrue(r[0])
        self.assertEqual(r[1].get("message"), "Fake Authentication")

        # We have an authentication, that is not read from authcache, since
        # the last_auth is too old = 10 minutes.
        delete_from_cache(username, realm, resolver, password)
        AuthCache(username, realm, resolver, pwd_hash,
                  first_auth=datetime.datetime.utcnow() - timedelta(hours=1),
                  last_auth=datetime.datetime.utcnow() - timedelta(
                      minutes=10)).save()
        r = auth_cache(fake_check_user_pass, User(username, realm),
                       password, options=options)
        self.assertTrue(r[0])
        self.assertEqual(r[1].get("message"), "Fake Authentication")

        # We have a policy, with no special last_auth
        delete_policy("pol1")
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   realm=realm,
                   resolver=resolver,
                   action="{0!s}={1!s}".format(ACTION.AUTH_CACHE, "4h"))
        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        delete_from_cache(username, realm, resolver, password)
        AuthCache(username, realm, resolver, pwd_hash,
                  first_auth=datetime.datetime.utcnow() - timedelta(hours=2),
                  last_auth=datetime.datetime.utcnow() - timedelta(
                      hours=1)).save()
        r = auth_cache(fake_check_user_pass, User(username, realm),
                       password, options=options)
        self.assertTrue(r[0])
        self.assertEqual(r[1].get("message"), "Authenticated by AuthCache.")

        # Test auth_cache policy with format "<seconds>/<#allowed authentications>"
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   realm=realm,
                   resolver=resolver,
                   action="{0!s}={1!s}".format(ACTION.AUTH_CACHE, "50s/2"))

        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}

        AuthCache(username, realm, resolver, pwd_hash,
                  first_auth=datetime.datetime.utcnow()).save()

        r = auth_cache(fake_check_user_pass, User(username, realm),
                       password, options=options)
        self.assertTrue(r[0])
        self.assertEqual("Authenticated by AuthCache.", r[1].get("message"))

        r = auth_cache(fake_check_user_pass, User(username, realm),
                       password, options=options)
        self.assertTrue(r[0])
        self.assertEqual("Authenticated by AuthCache.", r[1].get("message"))

        r = auth_cache(fake_check_user_pass, User(username, realm),
                       password, options=options)
        self.assertTrue(r[0])
        self.assertEqual("Fake Authentication", r[1].get("message"))

        delete_from_cache(username, realm, resolver, password)

        # Authentication not read from cache because first auth was too long ago
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   realm=realm,
                   resolver=resolver,
                   action="{0!s}={1!s}".format(ACTION.AUTH_CACHE, "50s/2"))

        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}

        AuthCache(username, realm, resolver, pwd_hash,
                  first_auth=datetime.datetime.utcnow() - timedelta(seconds=55)).save()

        r = auth_cache(fake_check_user_pass, User(username, realm),
                       password, options=options)
        self.assertTrue(r[0])
        self.assertEqual(r[1].get("message"), "Fake Authentication")

        # Clean up
        delete_policy("pol1")
        delete_realm(realm)
        delete_resolver(resolver)

    @radiusmock.activate
    def test_13_passthru_priorities(self):
        user = User("cornelius", realm="r1")
        passw = "test"
        options = {}
        # remove all tokens of cornelius
        remove_token(user=user)

        # A user with no tokens will fail to authenticate
        self.assertEqual(get_tokens(user=user, count=True), 0)
        rv = auth_user_passthru(check_user_pass, user, passw, options)
        self.assertFalse(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "The user has no tokens assigned")

        # Now set a PASSTHRU policy to the userstore
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   action="{0!s}=userstore".format(ACTION.PASSTHRU))
        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}
        rv = auth_user_passthru(check_user_pass, user, passw,
                                options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "against userstore due to 'pol1'")

        # Now add a PASSTHRU policy to a RADIUS config
        radiusmock.setdata(response=radiusmock.AccessAccept)
        set_policy(name="pol2",
                   scope=SCOPE.AUTH,
                   action="{0!s}=radiusconfig1".format(ACTION.PASSTHRU))
        r = add_radius("radiusconfig1", "1.2.3.4", "testing123",
                       dictionary=DICT_FILE)
        self.assertTrue(r > 0)
        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}

        # They will conflict, because they use the same priority
        with self.assertRaises(PolicyError):
            auth_user_passthru(check_user_pass, user, passw, options=options)

        # Lower pol1 priority
        set_policy(name="pol1", priority=2)

        rv = auth_user_passthru(check_user_pass, user, passw, options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "against RADIUS server radiusconfig1 due to 'pol2'")

        # Lower pol2 priority
        set_policy(name="pol2", priority=3)

        rv = auth_user_passthru(check_user_pass, user, passw, options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "against userstore due to 'pol1'")

        # Add old style priority
        set_policy(name="pol3",
                   scope=SCOPE.AUTH,
                   action=ACTION.PASSTHRU)

        rv = auth_user_passthru(check_user_pass, user, passw, options=options)
        self.assertTrue(rv[0])
        self.assertEqual(rv[1].get("message"),
                         "against userstore due to 'pol3'")
        set_policy(name="pol3", priority=2)

        # They will conflict, because they use the same priority
        with self.assertRaises(PolicyError):
            auth_user_passthru(check_user_pass, user, passw, options=options)

        delete_policy("pol3")

        # Now assign a token to the user. If the user has a token and the
        # passthru policy is set, the user must not be able to authenticate
        # with his userstore password.
        init_token({"serial": "PTHRU",
                    "type": "spass", "pin": "Hallo"},
                   user=user)
        rv = auth_user_passthru(check_user_pass, user, passw,
                                options=options)
        self.assertFalse(rv[0])
        self.assertEqual(rv[1].get("message"), "wrong otp pin")

        remove_token("PTHRU")
        delete_policy("pol1")
        delete_policy("pol2")

    def test_14_otppin_priority(self):
        my_user = User("cornelius", realm="r1")
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   action="{0!s}={1!s}".format(ACTION.OTPPIN, ACTIONVALUE.NONE),
                   priority=2)
        set_policy(name="pol2",
                   scope=SCOPE.AUTH,
                   action="{0!s}={1!s}".format(ACTION.OTPPIN, ACTIONVALUE.TOKENPIN),
                   priority=2)
        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        # error because of conflicting policies
        with self.assertRaises(PolicyError):
            auth_otppin(self.fake_check_otp, None,
                        "", options=options, user=my_user)

        # lower pol2 priority
        set_policy(name="pol2",
                   priority=3)

        # NONE with empty PIN -> success
        r = auth_otppin(self.fake_check_otp, None,
                        "", options=options, user=my_user)
        self.assertTrue(r)
        # NONE with empty PIN -> success, even if the authentication is done
        # for a serial and not a user, since the policy holds for all realms
        token = init_token({"type": "HOTP", "otpkey": "1234"})
        r = auth_otppin(self.fake_check_otp, token,
                        "", options=options, user=None)
        self.assertTrue(r)

        # NONE with some pin -> fail
        r = auth_otppin(self.fake_check_otp, None,
                        "some pin", options=options, user=my_user)
        self.assertFalse(r)

        # increase pol2 priority
        set_policy(name="pol2", priority=1)

        r = auth_otppin(self.fake_check_otp, None,
                        "FAKE", options=options,
                        user=my_user)
        self.assertTrue(r)
        r = auth_otppin(self.fake_check_otp, None,
                        "Wrong Pin", options=options,
                        user=my_user)
        self.assertFalse(r)
        delete_policy("pol1")
        delete_policy("pol2")

    def test_15_reset_all_failcounters(self):
        self.setUp_user_realms()

        set_policy("reset_all", scope=SCOPE.AUTH,
                   action=ACTION.RESETALLTOKENS)

        user = User(login="cornelius", realm=self.realm1)
        pin1 = "pin1"
        pin2 = "pin2"
        token1 = init_token({"serial": pin1, "pin": pin1,
                             "type": "spass"}, user=user)
        token2 = init_token({"serial": pin2, "pin": pin2,
                             "type": "spass"}, user=user)

        token1.inc_failcount()
        token2.inc_failcount()
        token2.inc_failcount()
        self.assertEqual(token1.token.failcount, 1)
        self.assertEqual(token2.token.failcount, 2)

        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        g.client_ip = None
        g.serial = None
        options = {"g": g}

        r = reset_all_user_tokens(self.fake_check_token_list,
                                  [token1, token2],
                                  "pw", None,
                                  options=options,
                                  allow_reset_all_tokens=True,
                                  result=True)
        self.assertTrue(r)

        self.assertEqual(token1.token.failcount, 0)
        self.assertEqual(token2.token.failcount, 0)

        # Clean up
        remove_token(pin1)
        remove_token(pin2)

    @radiusmock.activate
    def test_16_passthru_assign(self):
        user = User("cornelius", realm="r1")
        passw = "{0!s}test".format(self.valid_otp_values[1])
        options = {}
        # remove all tokens of cornelius
        remove_token(user=user)

        # create unassigned tokens in realm r1
        init_token({"type": "hotp",
                    "otpkey": "00" * 20,
                    "serial": "TOKFAIL"}, tokenrealms=["r1"])
        init_token({"type": "hotp",
                    "otpkey": self.otpkey,
                    "serial": "TOKMATCH"}, tokenrealms=["r1"])

        # A user with no tokens will fail to authenticate
        self.assertEqual(get_tokens(user=user, count=True), 0)
        rv = auth_user_passthru(check_user_pass, user, passw, options)
        self.assertFalse(rv[0])
        self.assertEqual(rv[1].get("message"), "The user has no tokens assigned")

        # Now add a PASSTHRU policy to a RADIUS config
        radiusmock.setdata(response=radiusmock.AccessAccept)
        set_policy(name="pol1",
                   scope=SCOPE.AUTH,
                   action="{0!s}=radiusconfig1".format(ACTION.PASSTHRU))
        r = add_radius("radiusconfig1", "1.2.3.4", "testing123",
                       dictionary=DICT_FILE)
        self.assertTrue(r > 0)
        set_policy(name="pol2",
                   scope=SCOPE.AUTH,
                   action="{0!s}=6:pin:1234".format(ACTION.PASSTHRU_ASSIGN))

        g = FakeFlaskG()
        g.policy_object = PolicyClass()
        g.audit_object = FakeAudit()
        options = {"g": g}

        rv = auth_user_passthru(check_user_pass, user, passw, options=options)
        self.assertTrue(rv[0])
        self.assertTrue("against RADIUS server radiusconfig1 due to 'pol1'" in rv[1].get("message"))
        self.assertIn("auto-assigned TOKMATCH", rv[1].get("message"))

        # Check if the token is assigned and can authenticate
        r = check_user_pass(User("cornelius", "r1"), "test{0!s}".format(self.valid_otp_values[2]))
        self.assertTrue(r[0])
        self.assertEqual(r[1].get("serial"), "TOKMATCH")

        remove_token("TOKFAIL")
        remove_token("TOKMATCH")
        delete_policy("pol1")
        delete_policy("pol2")

    def test_17_force_challenge_response(self):
        # If force_challenge_response is enabled, authentication will only be possible by sending the PIN in the first
        # step, then sending the OTP in the second step with a transaction_id returned in the first step.
        # Authenticating with PIN+OTP in one step will fail.
        self.setUp_user_realms()
        user = User(login="cornelius", realm=self.realm1)
        secret = "AAAAAAAAAA"
        pin = "1234"
        otps = ["984989", "457702", "527850", "820671", "569870"]
        token = init_token({"type": "hotp", "pin": pin, "otpkey": secret}, user=user)
        fake_g = FakeFlaskG()
        fake_g.policy_object = PolicyClass()
        fake_g.audit_object = FakeAudit()

        # Without any policies, the default PIN+OTP works
        res, reply_dict = check_user_pass(user, pin + otps[0], options={"g": fake_g})
        self.assertTrue(res)
        self.assertIn("message", reply_dict)
        self.assertIn("serial", reply_dict)
        self.assertIn("type", reply_dict)

        # With force_challenge_response policy, PIN+OTP will fail
        set_policy(name="force_cr", scope=SCOPE.AUTH, action=f"{ACTION.FORCE_CHALLENGE_RESPONSE}")
        # Also enable challenge-response for hotp type
        set_policy(name="hotp_cr", scope=SCOPE.AUTH, action=f"{ACTION.CHALLENGERESPONSE}=hotp")

        res, reply_dict = check_user_pass(user, pin + otps[1], options={"g": fake_g})
        self.assertFalse(res)
        self.assertIn("message", reply_dict)  # wrong otp pin message
        self.assertNotIn("serial", reply_dict)
        self.assertNotIn("type", reply_dict)

        # Using just the PIN will trigger the challenge
        res, reply_dict = check_user_pass(user, pin, options={"g": fake_g})
        self.assertIn("message", reply_dict)
        self.assertIn("serial", reply_dict)
        self.assertIn("type", reply_dict)
        self.assertIn("transaction_id", reply_dict)
        self.assertIn("multi_challenge", reply_dict)
        self.assertEqual(1, len(reply_dict["multi_challenge"]))

        # Using the OTP with the transaction_id will authenticate
        # Reuse the second OTP value because it was not consumed before
        res, reply_dict = check_user_pass(user, otps[1],
                                          options={"g": fake_g, "transaction_id": reply_dict["transaction_id"]})
        self.assertTrue(res)
        self.assertIn("message", reply_dict)
        self.assertIn("serial", reply_dict)
        # Cleanup
        delete_policy("force_cr")
        delete_policy("hotp_cr")
        user.delete()
        remove_token(token.token.serial)



"""
This file tests:

lib/eventhandler/logging.py
"""
from mock import mock
from datetime import datetime
from werkzeug.test import EnvironBuilder
from flask import Request, Response
from testfixtures import log_capture

from privacyidea.lib.token import init_token
from privacyidea.lib.eventhandler.logginghandler import LoggingEventHandler
from privacyidea.lib.user import User
from .base import MyTestCase, FakeFlaskG, FakeAudit


class LoggingTestCase(MyTestCase):
    def test_01_basefunctions(self):
        actions = LoggingEventHandler().actions
        self.assertIn('logging', actions, actions)

        # check positions
        pos = LoggingEventHandler().allowed_positions
        self.assertEqual(set(pos), {"post", "pre"}, pos)

    @log_capture()
    def test_02_loggingevent_default(self, capture):
        # simple logging event with default values
        g = FakeFlaskG()
        g.audit_object = FakeAudit()
        env = EnvironBuilder(method='POST', headers={}, path='/auth').get_environ()
        req = Request(env)
        req.all_data = {}
        resp = Response(response="""{"result": {"value": true}}""",
                        mimetype='application/json')
        options = {
            "g": g,
            "request": req,
            "response": resp,
            "handler_def": {
            }
        }
        log_handler = LoggingEventHandler()
        res = log_handler.do("logging", options=options)
        self.assertTrue(res)
        capture.check(
            ('pi-eventlogger', 'INFO', 'event=/auth triggered')
        )

    @log_capture()
    def test_03_loggingevent_parameter(self, capture):
        # simple logging event with user defined values
        self.setUp_user_realms()
        g = FakeFlaskG()
        g.audit_object = FakeAudit()
        env = EnvironBuilder(method='POST', headers={}, path='/auth').get_environ()
        req = Request(env)
        req.all_data = {}
        req.User = User("cornelius", self.realm1)
        resp = Response(response="""{"result": {"value": true}}""",
                        mimetype="application/json")
        options = {
            "g": g,
            "request": req,
            "response": resp,
            "handler_def": {
                'options': {
                    'name': 'eventlogger-privacyidea',
                    'level': 'WARN',
                    'message': 'Hello {username}!'
                }
            }
        }
        log_handler = LoggingEventHandler()
        res = log_handler.do("logging", options=options)
        self.assertTrue(res)
        capture.check_present(
            ('eventlogger-privacyidea', 'WARNING', 'Hello cornelius!')
        )

    @log_capture()
    def test_04_loggingevent_broken_parameter(self, capture):
        # simple logging event with faulty parameters
        self.setUp_user_realms()
        g = FakeFlaskG()
        g.audit_object = FakeAudit()
        env = EnvironBuilder(method='POST', headers={}, path='/auth').get_environ()
        req = Request(env)
        req.all_data = {}
        req.User = User("cornelius", self.realm1)
        resp = Response(response="""{"result": {"value": true}}""",
                        mimetype="application/json")
        options = {
            "g": g,
            "request": req,
            "response": resp,
            "handler_def": {
                'options': {
                    'name': None,
                    'level': 'some_level',
                    'message': None
                }
            }
        }
        log_handler = LoggingEventHandler()
        res = log_handler.do("logging", options=options)
        self.assertTrue(res)
        capture.check_present(
            ('root', 'INFO', 'event=/auth triggered')
        )

    @log_capture()
    def test_05_loggingevent_tags(self, capture):
        class UserAgentMock:
            string = "hello world"
            browser = "browser"

        # simple logging event with all tags
        self.setUp_sqlite_resolver_realm('testuser.sqlite', 'sqliterealm')
        available_tags = ['admin', 'realm', 'action', 'serial', 'url', 'user',
                          'surname', 'givenname', 'username', 'userrealm',
                          'tokentype', 'time', 'date', 'client_ip',
                          'ua_browser', 'ua_string']
        init_token({"serial": "testserial", "type": "spass",
                    "pin": "pin"}, user=User("cornelius", "sqliterealm"))
        g = FakeFlaskG()
        g.audit_object = FakeAudit()
        g.logged_in_user = {"username": "admin", "role": "admin",
                            "realm": "super"}
        env = EnvironBuilder(method='POST', headers={}, path='/auth').get_environ()
        req = Request(env)
        req.user_agent = UserAgentMock()
        req.all_data = {'serial': 'testserial'}
        req.User = User("cornelius", 'sqliterealm')
        resp = Response(response="""{"result": {"value": true}}""",
                        mimetype="application/json")
        options = {
            "g": g,
            "request": req,
            "response": resp,
            "handler_def": {
                'options': {
                    'message': ' '.join(['{0!s}={{{0!s}}}'.format(x) for x in available_tags])
                }
            }
        }
        current_utc_time = datetime(2018, 3, 4, 5, 6, 8)
        with mock.patch('privacyidea.lib.utils.datetime') as mock_dt:
            mock_dt.now.return_value = current_utc_time

            log_handler = LoggingEventHandler()
            res = log_handler.do("logging", options=options)
            self.assertTrue(res)
            capture.check_present(
                ('pi-eventlogger', 'INFO',
                 'admin=admin realm=super action=/auth serial=testserial '
                 'url=http://localhost/ user=Cornelius surname=Kölbel '
                 'givenname=Cornelius username=cornelius userrealm=sqliterealm '
                 'tokentype=spass time=05:06:08 date=2018-03-04 '
                 'client_ip=None ua_browser=browser ua_string=hello world')
            )



#   2016-06-15 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#              Add allowed parameters to the SMS Provider
#
#    privacyIDEA is a fork of LinOTP
#    May 28, 2014 Cornelius Kölbel
#    E-mail: info@privacyidea.org
#    Contact: www.privacyidea.org
#
#    2015-01-30 Rewrite for flask migration
#               Cornelius Kölbel <cornelius@privacyidea.org>
#
#
#    Copyright (C) LinOTP: 2010 - 2014 LSE Leading Security Experts GmbH
#
#    This program is free software: you can redistribute it and/or
#    modify it under the terms of the GNU Affero General Public
#    License, version 3, as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the
#               GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__doc__="""This is the SMSClass to send SMS via SMTP Gateway.
i.e. a Mail is sent to an Gateway/Emailserver and dependig on the
address, subject and body this gateway will trigger the sending of the SMS.

The code is tested in tests/test_lib_smsprovider
"""
from privacyidea.lib.smsprovider.SMSProvider import ISMSProvider, SMSError
from privacyidea.lib.smtpserver import send_email_identifier, send_email_data
from privacyidea.lib import _
import logging
log = logging.getLogger(__name__)

PHONE_TAG = "<phone>"
MSG_TAG = "<otp>"


class SmtpSMSProvider(ISMSProvider):

    def submit_message(self, phone, message):
        """
        Submits the message for phone to the email gateway.

        Returns true in case of success

        In case of a failure an exception is raised
        """
        if self.smsgateway:
            phone = self._mangle_phone(phone, self.smsgateway.option_dict)
            identifier = self.smsgateway.option_dict.get("SMTPIDENTIFIER")
            recipient = self.smsgateway.option_dict.get("MAILTO").format(
                otp=message, phone=phone)
            subject = self.smsgateway.option_dict.get("SUBJECT",
                                                      "{phone}").format(
                otp=message, phone=phone)
            body = self.smsgateway.option_dict.get("BODY", "{otp}").format(
                otp=message, phone=phone)
        else:
            phone = self._mangle_phone(phone, self.config)
            identifier = self.config.get("IDENTIFIER")
            server = self.config.get("MAILSERVER")
            sender = self.config.get("MAILSENDER")
            recipient = self.config.get("MAILTO")
            subject = self.config.get("SUBJECT", PHONE_TAG)
            body = self.config.get("BODY", MSG_TAG)

            if not (server and recipient and sender) and not (identifier and \
                    recipient):
                log.error("incomplete config: %s. MAILTO and (IDENTIFIER or "
                          "MAILSERVER and MAILSENDER) needed" % self.config)
                raise SMSError(-1, "Incomplete SMS config.")

            recipient = recipient.replace(PHONE_TAG, phone)
            subject = subject.replace(PHONE_TAG, phone)
            subject = subject.replace(MSG_TAG, message)
            body = body.replace(PHONE_TAG, phone)
            body = body.replace(MSG_TAG, message)

        log.debug("submitting message {0!r} to {1!s}".format(body, phone))
        if identifier:
            r = send_email_identifier(identifier, recipient, subject, body)
        else:
            username = self.config.get("MAILUSER")
            password = self.config.get("MAILPASSWORD")
            r = send_email_data(server, subject, body, sender, recipient,
                                username, password)
        if not r:
            raise SMSError(500, "Failed to deliver SMS to SMTP Gateway.")

        return True

    @classmethod
    def parameters(cls):
        """
        Return a dictionary, that describes the parameters and options for the
        SMS provider.
        Parameters are required keys to values.

        :return: dict
        """
        from privacyidea.lib.smtpserver import get_smtpservers
        params = {"options_allowed": False,
                  "headers_allowed": False,
                  "parameters": {
                      "MAILTO": {
                          "required": True,
                          "description": "The recipient of the email. "
                                         "Use tags {phone} and {otp}."},
                      "SMTPIDENTIFIER": {
                          "required": True,
                          "description": "Your SMTP configuration, "
                                         "that should be used to send the "
                                         "email.",
                          "values": [
                              provider.config.identifier for
                              provider in get_smtpservers()]},
                      "SUBJECT": {
                          "description": "The optional subject of the email. "
                                         "Use tags {phone} and {otp}."},
                      "BODY": {
                          "description": "The optional body of the email. "
                                         "Use tags {phone} and {otp}.",
                          "type": "text"},
                      "REGEXP": {
                          "description": cls.regexp_description
                      }
                  }
                  }
        return params



# SPDX-FileCopyrightText: (C) 2023 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-FileCopyrightText: (C) 2014 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# 2024-04-22 Paul Lettich <paul.lettich@netknights.it>
#            Refactor using click
# 2020-11-18 Henning Hollermann <henning.hollermann@netknights.it>
#            Allow import and export of events, resolvers and policies
# 2018-08-07 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Allow creation of HSM keys
# 2017-10-08 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Allow cleaning up different actions with different
#            retention times.
# 2017-07-12 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add generation of PGP keys
# 2017-02-23 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add CA sub commands
# 2017-01-27 Diogenes S. Jesus
#            Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add creation of more detailed policy
# 2016-04-15 Cornelius Kölbel <cornelius@privacyidea.org>
#            Add backup for pymysql driver
# 2016-01-29 Cornelius Kölbel <cornelius@privacyidea.org>
#            Add profiling
# 2015-10-09 Cornelius Kölbel <cornelius@privacyidea.org>
#            Set file permissions
# 2015-09-24 Cornelius Kölbel <cornelius@privacyidea.org>
#            Add validate call
# 2015-06-16 Cornelius Kölbel <cornelius@privacyidea.org>
#            Add creation of JWT token
# 2015-03-27 Cornelius Kölbel, cornelius@privacyidea.org
#            Add sub command for policies
# 2014-12-15 Cornelius Kölbel, info@privacyidea.org
#            Initial creation
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""CLI Tool for configuring and managing privacyIDEA"""
import click
import copy
from flask.cli import FlaskGroup, run_command
from privacyidea.cli import create_silent_app, get_version
from privacyidea.lib.utils import get_version_number
from .admin import admin_cli
from .audit import audit_cli, rotate_audit as audit_rotate_audit
from .backup import backup_cli
from .pi_setup import (setup_cli, encrypt_enckey, create_enckey, create_tables,
                       create_pgp_keys, create_audit_keys, drop_tables)
from .pi_config import (config_cli, ca_cli, realm_cli, resolver_cli, event_cli,
                        policy_cli, authcache_cli, hsm_cli)
from .api import api_cli
from .token import token_cli

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(cls=FlaskGroup, create_app=create_silent_app, context_settings=CONTEXT_SETTINGS,
             add_version_option=False,
             epilog='Check out our docs at https://privacyidea.readthedocs.io/ for more details')
@click.option(
    "--version",
    help="Show the version information",
    expose_value=False,
    callback=get_version,
    is_flag=True,
    is_eager=True,
)
def cli():
    """
\b
             _                    _______  _______
   ___  ____(_)  _____ _______ __/  _/ _ \\/ __/ _ |
  / _ \\/ __/ / |/ / _ `/ __/ // // // // / _// __ |
 / .__/_/ /_/|___/\\_,_/\\__/\\_, /___/____/___/_/ |_|
/_/                       /___/

  Management script for the privacyIDEA application."""
    click.echo(r"""
             _                    _______  _______
   ___  ____(_)  _____ _______ __/  _/ _ \/ __/ _ |
  / _ \/ __/ / |/ / _ `/ __/ // // // // / _// __ |
 / .__/_/ /_/|___/\_,_/\__/\_, /___/____/___/_/ |_|
/_/                       /___/
{0!s:>51}
    """.format('v{0!s}'.format(get_version_number())), err=True)


deprecated_commands = [
    (audit_rotate_audit, "rotate_audit", "audit rotate"),
    (create_enckey, None, "setup create_enckey"),
    (encrypt_enckey, None, "setup encrypt_enckey"),
    (create_audit_keys, None, "setup create_audit_keys"),
    (create_tables, None, "setup create_tables"),
    (create_pgp_keys, None, "setup create_pgp_keys"),
    (create_tables, "createdb", "setup create_tables"),
    (drop_tables, None, "setup drop_tables"),
    (drop_tables, "dropdb", "setup  drop_tables"),
    (realm_cli, None, "config realm"),
    (resolver_cli, None, "config resolver"),
    (policy_cli, None, "config policy"),
    (event_cli, None, "config event"),
    (ca_cli, None, "config ca"),
    (authcache_cli, None, "config authcache"),
    (hsm_cli, None, "config hsm"),
    (run_command, "runserver", "run")
]

for cmd, new_name, epilog in deprecated_commands:
    new_cmd = copy.copy(cmd)
    new_cmd.deprecated = True
    new_cmd.hidden = True
    if new_name:
        new_cmd.name = new_name
    new_cmd.epilog = f"This command is deprecated. Please use 'pi-manage {epilog}' instead."
    cli.add_command(new_cmd)

cli.add_command(admin_cli)
cli.add_command(audit_cli)
cli.add_command(setup_cli)
cli.add_command(config_cli)
cli.add_command(backup_cli)
cli.add_command(api_cli)
cli.add_command(token_cli)

if __name__ == '__main__':
    cli()



# SPDX-FileCopyrightText: (C) 2024 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.

import pytest
from sqlalchemy.orm.session import close_all_sessions

from privacyidea.app import create_app
from privacyidea.models import db
from privacyidea.lib.lifecycle import call_finalizers
from privacyidea.cli.tools.cron import cli as privacyidea_cron


@pytest.fixture(scope="class")
def app():
    """Create and configure app instance for testing"""
    app = create_app(config_name="testing", config_file="", silent=True)
    with app.app_context():
        db.create_all()

    yield app

    with app.app_context():
        call_finalizers()
        close_all_sessions()
        db.drop_all()
        db.engine.dispose()


class TestPICronExec:
    def test_01_picron_exec(self, app):
        runner = app.test_cli_runner()
        result = runner.invoke(privacyidea_cron, [])
        assert "Usage: cli [OPTIONS] COMMAND [ARGS]..." in result.output, result
        assert "Execute all periodic tasks that are scheduled to run." in result.output, result
        assert "run_scheduled" in result.output, result
        assert "Show a list of available tasks that could be run." in result.output, result
        assert "Manually run a periodic task" in result.output, result
        assert "run_manually" in result.output, result

        result = runner.invoke(privacyidea_cron, ["list"])
        assert "Active  ID  Name" in result.output, result.output

        result = runner.invoke(privacyidea_cron, ["run_scheduled"])
        assert "There are no tasks scheduled on node Node1." in result.output, result.output

        result = runner.invoke(privacyidea_cron, ["run_scheduled", "-c"])
        assert not result.output, result.output



# http://www.privacyidea.org
# (c) cornelius kölbel, privacyidea.org
#
# 2014-12-08 Cornelius Kölbel, <cornelius@privacyidea.org>
#            Complete rewrite during flask migration
#            Try to provide REST API
#
# privacyIDEA is a fork of LinOTP. Some code is adapted from
# the system-controller from LinOTP, which is
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """The realm endpoints are used to define realms.
A realm groups together many users. Administrators can manage the tokens of
the users in such a realm. Policies and tokens can be assigned to realms.

A realm consists of several resolvers. Thus you can create a realm and gather
users from LDAP and flat file source into one realm or you can pick resolvers
that collect users from different points from your vast LDAP directory and
group these users into a realm.

You will only be able to see and use user object, that are contained in a realm.

The code of this module is tested in tests/test_api_system.py
"""
from flask import Blueprint, request, current_app, g
from .lib.utils import (getParam,
                        required,
                        send_result,
                        get_priority_from_param)
from ..lib.log import log_with
from ..lib.realm import (set_default_realm,
                         get_default_realm,
                         set_realm,
                         get_realms,
                         delete_realm)
from ..api.lib.prepolicy import prepolicy, check_base_action
from ..lib.utils import reduce_realms
from privacyidea.lib import _
from privacyidea.lib.auth import ROLE
from privacyidea.lib.config import check_node_uuid_exists
from privacyidea.lib.error import ParameterError
from privacyidea.lib.policy import CONDITION_CHECK, ACTION, Match
import logging

log = logging.getLogger(__name__)

realm_blueprint = Blueprint('realm_blueprint', __name__)
defaultrealm_blueprint = Blueprint('defaultrealm_blueprint', __name__)


# ----------------------------------------------------------------
#
#  REALM functions
#
#

@realm_blueprint.route('/<realm>', methods=['POST'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.RESOLVERWRITE)
def set_realm_api(realm=None):
    """
    This call creates a new realm or reconfigures a realm.
    The realm contains a list of resolvers.

    In the result it returns a list of added resolvers and a list of
    resolvers, that could not be added.

    :arg realm: The unique name of the realm
    :<json string/list resolvers: A comma separated list of unique resolver
        names or a list object
    :<json integer priority: Additional priority parameters ``priority.<resolvername>``
        to define the priority of the resolvers within this realm
    :>json bool status: Status of the request
    :>json value: object with a list of added and failed resolvers
    :reqheader PI-Authorization: The authorization token

    **Example request**:

    To create a new realm ``newrealm``, that consists of the resolvers
    ``reso1_with_realm`` and ``reso2_with_realm`` call:

    .. sourcecode:: http

       POST /realm/newrealm HTTP/1.1
       Host: example.com
       Accept: application/json
       Content-Type: application/json

       "resolvers": "reso1_with_realm, reso2_with_realm"
       "priority.reso1_with_realm": 1
       "priority.reso2_with_realm": 2

    **Example response**:

    .. sourcecode:: http

       HTTP/1.1 200 OK
       Content-Type: application/json

       {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
                    "status": true,
                    "value": {
                        "added": ["reso1_with_realm", "reso2_with_realm"],
                        "failed": []
                    }
          }
          "version": "privacyIDEA unknown"
       }
    """
    param = request.all_data
    resolvers = getParam(param, "resolvers", required)
    priority = get_priority_from_param(param)

    if isinstance(resolvers, list):
        resolver_list = resolvers
    else:
        resolver_list = resolvers.split(',')
    resolvers = [{'name': res,
                  'priority': priority.get(res, None)
                  } for res in resolver_list]
    # since this endpoint is not node-specific, we would delete all other resolvers
    # which are specified for a different node.
    # We need to add all the other resolvers to the dictionary to avoid deleting them
    orig_resolvers = get_realms(realm).get(realm, {}).get("resolver", [])
    for res in orig_resolvers:
        if res.get("node", ""):
            resolvers.append({'name': res["name"],
                              'priority': res.get("priority"),
                              "node": res.get("node")})
    (added, failed) = set_realm(realm, resolvers=resolvers)
    g.audit_object.log({'success': not failed,
                        'info':  "realm: {0!r}, resolvers: {1!r}".format(realm,
                                                                         resolvers)})
    return send_result({"added": added,
                        "failed": failed})


@realm_blueprint.route('/', methods=['GET'])
@log_with(log)
def get_realms_api():
    """
    This call returns the list of all defined realms.
    It takes no arguments.

    :return: a json result with a list of realms
    :reqheader PI-Authorization: The authorization token

    **Example request**:

    .. sourcecode:: http

       GET /realm/ HTTP/1.1
       Host: example.com
       Accept: application/json

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": {
              "realm1_with_resolver": {
                "default": true,
                "resolver": [
                  {
                    "name": "reso1_with_realm",
                    "node": "",
                    "priority": null,
                    "type": "passwdresolver"
                  }
                ]
              }
            }
          },
          "version": "privacyIDEA unknown"
        }

    .. versionchanged:: 3.10 The response contains the node and priority of the resolver
    """
    all_realms = get_realms()
    g.audit_object.log({"success": True})
    # This endpoint is called by admins anyway
    luser = g.logged_in_user
    policies = Match.generic(g, scope=luser.get("role", ROLE.ADMIN),
                             adminrealm=luser.get("realm"),
                             adminuser=luser.get("username"),
                             active=True,
                             extended_condition_check=CONDITION_CHECK.DO_NOT_CHECK_AT_ALL).policies()
    realms = reduce_realms(all_realms, policies)

    return send_result(realms)


@realm_blueprint.route('/superuser', methods=['GET'])
@log_with(log)
def get_super_user_realms():
    """
    This call returns the list of all superuser realms
    as they are defined in *pi.cfg*.
    See :ref:`cfgfile` for more information about this.

    :return: a json result with a list of superuser realms
    :reqheader PI-Authorization: The authorization token

    **Example request**:

    .. sourcecode:: http

       GET /realm/superuser HTTP/1.1
       Host: example.com
       Accept: application/json

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": [
              "superuser",
              "realm2"
            ]
          },
          "version": "privacyIDEA unknown"
        }
    """
    superuser_realms = current_app.config.get("SUPERUSER_REALM", [])
    g.audit_object.log({"success": True})
    return send_result(superuser_realms)


@defaultrealm_blueprint.route('/<realm>', methods=['POST'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.RESOLVERWRITE)
def set_default_realm_api(realm=None):
    """
    This call sets the default realm.

    :param realm: the name of the realm, that should be the default realm
    :>json bool status: Status of the request
    :>json integer value: The id of the new default realm
    :reqheader PI-Authorization: The authorization token

    **Example request**:

    .. sourcecode:: http

       POST /defaultrealm/new_default_realm HTTP/1.1
       Host: example.com

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": 1
          },
          "version": "privacyIDEA unknown"
        }
    """
    realm = realm.lower().strip()
    r = set_default_realm(realm)
    g.audit_object.log({"success": r,
                        "info": realm})
    return send_result(r)


@defaultrealm_blueprint.route('', methods=['DELETE'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.RESOLVERDELETE)
def delete_default_realm_api(realm=None):
    """
    This call deletes the default realm.

    :>json bool status: Status of the request
    :>json integer value: The id of the realm which used to be default
        or ``0`` in case no default realm was found
    :reqheader PI-Authorization: The authorization token

    **Example request**:

    .. sourcecode:: http

       DELETE /defaultrealm HTTP/1.1
       Host: example.com

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": 1
          },
          "version": "privacyIDEA unknown"
        }
    """
    r = set_default_realm("")
    g.audit_object.log({"success": True,
                        "info": ""})
    return send_result(r)


@defaultrealm_blueprint.route('', methods=['GET'])
@log_with(log)
def get_default_realm_api():
    """
    This call returns the default realm

    :return: a json description of the default realm with the resolvers
    :reqheader PI-Authorization: The authorization token

    **Example request**:

    .. sourcecode:: http

       GET /defaultrealm HTTP/1.1
       Host: example.com

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": {
              "defrealm": {
                "default": true,
                "id": 1,
                "resolver": [
                  {
                    "name": "defresolver",
                    "node": "",
                    "priority": null,
                    "type": "passwdresolver"
                  }
                ]
              }
            }
          },
          "version": "privacyIDEA unknown"
        }
    """
    res = {}
    defRealm = get_default_realm()
    if defRealm:
        res = get_realms(defRealm)

    g.audit_object.log({"success": True,
                        "info": defRealm})

    return send_result(res)


@realm_blueprint.route('/<realm>', methods=['DELETE'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.RESOLVERDELETE)
def delete_realm_api(realm=None):
    """
    This call deletes the given realm.

    :param realm: The name of the realm to delete
    :>json bool status: Status of the request
    :>json integer value: The id of the deleted realm
    :reqheader PI-Authorization: The authorization token

    **Example request**:

    .. sourcecode:: http

       DELETE /realm/realm_to_delete HTTP/1.1
       Host: example.com
       Accept: application/json

    **Example response**:

    .. sourcecode:: http

       HTTP/1.1 200 OK
       Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": 1
          },
          "version": "privacyIDEA unknown"
        }
    """
    ret = delete_realm(realm)
    g.audit_object.log({"success": ret > 0,
                        "info": realm})

    return send_result(ret)


# TODO: Flask supports conversion of the nodeid to a UUID, but in get_all_params()
#  we put everything (including the viewargs) in the param dict and try to "unquote"
#  all entries. This fails for the UUID type. We should probably remove the viewargs from
#  the param dict and handle them separately in the corresponding API function.
@realm_blueprint.route('/<string:realm>/node/<string:nodeid>', methods=['POST'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.RESOLVERWRITE)
def set_realm_node_api(realm, nodeid):
    """
    This call creates or reconfigures a realm with node specific settings.

    The realm contains a list of resolvers and corresponding priorities per node.
    If the node UUID can not be found in the database, the realm will be
    configured without a specific node.
    In the result it returns a list of added resolvers and a list of
    resolvers, that could not be added.

    :arg string realm: The unique name of the realm
    :arg string nodeid: The UUID of the node
    :<json resolver: A JSON object with a list consisting of objects with the
      resolver name and priority
    :reqheader PI-Authorization: The authorization token
    :statuscode 200: no error
    :statuscode 400:
      - The given node UUID does not exist
      - Could not verify data in request

    **Example request**:

    Create a new realm ``newrealm``, that consists of the resolvers
    ``resolver1`` and ``resolver2`` on the node with
    UUID ``8e4272a9-9037-40df-8aa3-976e4a04b5a9``:

    .. sourcecode:: http

      POST /realm/newrealm/node/8e4272a9-9037-40df-8aa3-976e4a04b5a9 HTTP/1.1
      Host: example.com
      Accept: application/json
      Content-Type: application/json

      {
        "resolver": [
          {
            "name": "resolver1",
            "priority": 1
          },
          {
            "name": "resolver2",
            "priority": 2
          }
        ]
      }

    **Example response**:

    .. sourcecode:: http

       HTTP/1.1 200 OK
       Content-Type: application/json

       {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
                    "status": true,
                    "value": {
                        "added": ["resolver1", "resolver2"],
                        "failed": []
                    }
          }
          "version": "privacyIDEA unknown"
       }

    .. versionadded:: 3.10 Node specific realm configuration
    """
    if not check_node_uuid_exists(nodeid):
        log.warning(f"Node with UUID {nodeid} does not exist in the database!")
        raise ParameterError(_("The given node does not exist!"))

    data = request.json
    resolvers = []
    if "resolver" not in data:
        log.warning("Missing resolver data in request")
        raise ParameterError(_("Could not verify data in request!"))

    for res in data["resolver"]:
        try:
            resolvers.append({"name": res["name"],
                              "priority": int(res.get("priority")) if res.get("priority") else None,
                              "node": nodeid})
        except (KeyError, ValueError) as e:
            log.warning(f"Could not parse resolver data {res}: {e}")
            log.debug(e.__traceback__)
            raise ParameterError(_("Could not verify data in request!"))

    # since this endpoint is node-specific, we would delete all other resolvers
    # which are specified for a different node or no node at all. We need to
    # add all the other resolvers to the dictionary to avoid deleting them
    orig_resolvers = get_realms(realm).get(realm, {}).get("resolver", [])
    for res in orig_resolvers:
        if res.get("node", "") != nodeid:
            resolvers.append({'name': res["name"],
                              'priority': res.get("priority"),
                              "node": res.get("node")})
    (added, failed) = set_realm(realm, resolvers=resolvers)

    g.audit_object.log({
        'success': not failed,
        # Overwrite resolver entry in audit log since `before_after` added a dict
        'resolver': ", ".join([r["name"] for r in resolvers]),
        'info': "realm: {0!r}, resolvers: {1!r}".format(realm,
                                                        [r["name"] for r in resolvers])})
    return send_result({"added": added,
                        "failed": failed})



"""v3.7: Allow empty machine_id and resolver_id in machinetoken

Revision ID: ef29ba43e290
Revises: ff26585932ec
Create Date: 2021-12-30 12:17:15.521336

"""

# revision identifiers, used by Alembic.
revision = 'ef29ba43e290'
down_revision = 'ff26585932ec'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.alter_column('machinetoken', 'machine_id',
                   existing_type=sa.VARCHAR(length=255),
                   nullable=True)
        op.alter_column('machinetoken', 'machineresolver_id',
                   existing_type=sa.INTEGER(),
                   nullable=True)
    except Exception as exx:
        print("Failed to make machine_id and resolver_id in machinetokens nullable.")
        print(exx)


def downgrade():
    try:
        op.alter_column('machinetoken', 'machineresolver_id',
                   existing_type=sa.INTEGER(),
                   nullable=False)
        op.alter_column('machinetoken', 'machine_id',
                   existing_type=sa.VARCHAR(length=255),
                   nullable=False)
    except Exception as exx:
        print("Failed to make machine_id and resolver_id in machinetokens non-nullable.")
        print(exx)



#!/usr/bin/env python
#  2020-06-13 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add failsafe for tokeninfo
#  2019-10-04 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Allow encryption and config file
#  2018-02-09 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Speedup thanks to Friedrich Weber.
#             Add flexible resolver mapping
#  2016-09-14 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
__doc__ = """You can use this script to migrate a LinOTP database to privacyIDEA.
All tokens and token assignments are transferred to privacyIDEA.

You only need an export of the LinOTP "Token" table.
You need to generate a config file and adapt it to your needs.

SQL section contains the database connections to the LinOTP and
the privacyIDEA database.

SQL:INSERT_CHUNK_SIZE gives you a speed advantage. This many tokens will
be inserted at one time.

MIGRATE defines which elements should be migrated: This can be tokens, tokeninfo and assignment.

The ASSIGNMENTS section defines, to which privacyIDEA users the old token-assignments 
    should be migrated to.

The "resolver" key maps LinOTP-resolvers to privacyIDEA-Resolvers. 
The key is the LinOTP resolver and the value is the new privacyIDEA resolver.

The "realm" key puts tokens that are migrated into the specified privacyIDEA resolver (key)
into the given privacyIDEA realm (value). Note: There is no logic checking with the 
privacyIDEA resolver-realm-configuration done.
So take care that the privacyIDEA resolver is really located in the specified privacyIDEA realm.

In this example the resolvername "lokal" from LinOTP gets imported to
privacyIDEA with the resolver "PIResolver" in realm "pirealm".

Tokens, that were not assigned to a user will be assigned to the realms specified
in the list "unassigned_tokens".

LinOTP (sometimes) uses a mixed endian notation for the objectGUID of users in Active Directory.
So if your users can not be found due to an unknown objectGUID, you need to set "convert_endian"
to true. This will convert the mixed endian notation to a standard notation used by privacyIDEA.

The NEW_TOKENINFO section can set any arbitrary tokeninfo for the 
migrated tokens.
"""
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import select
from sqlalchemy.schema import ForeignKey, Sequence
from sqlalchemy import (Table, MetaData, Column, Integer, Unicode, Boolean,
                        UnicodeText)
import sys
import json
import getopt
import binascii
import os
import re
from privacyidea.lib.crypto import aes_cbc_decrypt, aes_cbc_encrypt
from privacyidea.lib.utils import hexlify_and_unicode


EXAMPLE_CONFIG_FILE = """{
    "SQL": {
        "LINOTP_URI": "mysql+pymysql://linotp:linotp@localhost/linotp2",
        "PRIVACYIDEA_URI": "mysql+pymysql://pi:QEK_k_f3SGbu@localhost/pi?charset=utf8mb4",
        "INSERT_CHUNK_SIZE": 10000
    },
    "MIGRATE": {
        "tokens": true,
        "tokeninfo": true,
        "tokeninfo_unicode_keys": ["phone"],
        "assignments": false
    },
    "ASSIGNMENTS": {
        "resolver": {"local": "localpw"},
        "realm": {"localpw": "neuerrealm"},
        "unassigned_tokens": ["testfoo"],
        "convert_endian": true
    },
    "ENCRYPTION": {
        "reencrypt": false,
        "linotp_enc_file": "/etc/linotp2/encKey",
        "privacyidea_enc_file": "/etc/privacyidea/enckey"
    },
    "NEW_TOKENINFO": {
        "arbitrary_migrationkey": "somevalue"
    }
}"""


# These patterns are used to convert a middle endian UUID to a standard UUID:
# 33221100554477668899AABBCCDDEEFF -> 00112233-4455-6677-8899-AABBCCDDEEFF
UUID_MATCH_PATTERN = "^([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})" \
                     "([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})" \
                     "([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})" \
                     "([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$"

UUID_REPLACE_PATTERN = "\\4\\3\\2\\1-\\6\\5-\\8\\7-\\9\\10-\\11\\12\\13\\14\\15\\16"


class Config(object):

    def __init__(self, config_file):
        with open(config_file, "r") as f:
            contents = f.read()
        config = json.loads(contents)
        self.ASSIGNMENTS = config.get("ASSIGNMENTS")
        self.LINOTP_URI = config.get("SQL").get("LINOTP_URI")
        self.PRIVACYIDEA_URI = config.get("SQL").get("PRIVACYIDEA_URI")
        self.INSERT_CHUNK_SIZE = config.get("SQL").get("INSERT_CHUNK_SIZE")
        self.MIGRATE = config.get("MIGRATE")
        self.REENCRYPT = config.get("ENCRYPTION", {}).get("reencrypt")
        self.LINOTP_ENC_FILE = config.get("ENCRYPTION", {}).get("linotp_enc_file")
        self.PRIVACYIDEA_ENC_FILE = config.get("ENCRYPTION", {}).get("privacyidea_enc_file")
        self.NEW_TOKENINFO = config.get("NEW_TOKENINFO", {})

        if self.REENCRYPT:
            self.LINOTP_ENC_KEY = read_enckey(self.LINOTP_ENC_FILE)
            self.PRIVACYIDEA_ENC_KEY = read_enckey(self.PRIVACYIDEA_ENC_FILE)
            if not (self.LINOTP_ENC_KEY and self.PRIVACYIDEA_ENC_KEY):
                print("Failed to read both enckey files.")
                sys.exit(2)
        else:
            self.LINOTP_ENC_KEY = self.PRIVACYIDEA_ENC_KEY = None


def reencrypt(enc_data, iv, old_key, new_key):
    # decrypt LinOTP
    iv = binascii.unhexlify(iv)
    input_data = binascii.unhexlify(enc_data)
    output = aes_cbc_decrypt(old_key, iv, input_data)
    if not output:
        raise Exception('invalid encoded secret!')
    # unpad
    output = output.rstrip(b'\0')
    # The stripping of the ascii chars was missing for LinOTP 2.7.2
    eof = output.rfind(b"\x01\x02")
    if eof >= 0:
        output = output[:eof]
    data = binascii.a2b_hex(output)

    # encrypt anew
    iv = os.urandom(16)
    input_data = binascii.b2a_hex(data)
    input_data += b"\x01\x02"
    padding = (16 - len(input_data) % 16) % 16
    input_data += padding * b"\0"

    res = aes_cbc_encrypt(new_key, iv, input_data)
    return hexlify_and_unicode(res), hexlify_and_unicode(iv)


def dict_without_keys(d, keys):
    new_d = d.copy()
    for key in keys:
        if key in d:
            new_d.pop(key)
    return new_d


def migrate(config_obj):
    print("Re-Encryption token data: {0!s}".format(config_obj.REENCRYPT))

    # This maps the resolver types. You must not change this!
    resolver_map = {"LDAPIdResolver": "ldapresolver",
                    "SQLIdResolver": "sqlresolver",
                    "PasswdIdResolver": "passwdresolver"}

    token_serial_id_map = {}
    realm_id_map = {}
    resolver_id_map = {}

    metadata = MetaData()

    token_table = Table("token", metadata,
                        Column("id", Integer, Sequence('token_seq'),
                               primary_key=True, nullable=False),
                        Column("description", Unicode(80), default=''),
                        Column("serial", Unicode(40), default='', unique=True,
                               nullable=False, index=True),
                        Column("tokentype", Unicode(30), default='HOTP',
                               index=True),
                        Column("user_pin", Unicode(512), default=''),
                        Column("user_pin_iv", Unicode(32), default=''),
                        Column("so_pin", Unicode(512), default=''),
                        Column("so_pin_iv", Unicode(32), default=''),
                        Column("pin_seed", Unicode(32), default=''),
                        Column("otplen", Integer(), default=6),
                        Column("pin_hash", Unicode(512), default=''),
                        Column("key_enc", Unicode(1024), default=''),
                        Column("key_iv", Unicode(32), default=''),
                        Column("maxfail", Integer(), default=10),
                        Column("active", Boolean(), nullable=False, default=True),
                        Column("revoked", Boolean(), default=False),
                        Column("locked", Boolean(), default=False),
                        Column("failcount", Integer(), default=0),
                        Column("count", Integer(), default=0),
                        Column("count_window", Integer(), default=10),
                        Column("sync_window", Integer(), default=1000),
                        Column("rollout_state", Unicode(10), default=''))

    tokenowner_table = Table("tokenowner", metadata,
                             Column("id", Integer, Sequence('tokenowner_seq'),
                                    primary_key=True, nullable=False),
                             Column("token_id", Integer, ForeignKey("token.id"), nullable=False),
                             Column("resolver", Unicode(120), default="", index=True),
                             Column("user_id", Unicode(320), default="", index=True),
                             Column("realm_id", Integer, ForeignKey("realm.id"), nullable=False)
                             )

    tokeninfo_table = Table("tokeninfo", metadata,
                            Column("id", Integer, Sequence('tokeninfo_seq'),
                                   primary_key=True),
                            Column("Key", Unicode(255), nullable=False),
                            Column("Value", UnicodeText(), default=''),
                            Column("Type", Unicode(100), default=''),
                            Column("Description", Unicode(2000), default=''),
                            Column("token_id", Integer()))

    tokenrealm_table = Table("tokenrealm", metadata,
                             Column("id", Integer(),
                                    Sequence('tokenrealm_seq'),
                                    primary_key=True),
                             Column("token_id", Integer()),
                             Column("realm_id", Integer()))

    realm_table = Table("realm", metadata,
                        Column("id", Integer, Sequence('realm_seq'),
                               primary_key=True),
                        Column("name", Unicode(255), default=''),
                        Column("default",  Boolean(), default=False),
                        Column("option", Unicode(40), default=''))

    resolver_table = Table("resolver", metadata,
                           Column("id", Integer, Sequence('resolver_seq'),
                                  primary_key=True),
                           Column("name", Unicode(255), default=""),
                           Column("rtype", Unicode(255), default=""))

    resolver_config_table = Table("resolverconfig", metadata,
                                  Column("id", Integer,
                                         Sequence('resolverconf_seq'),
                                         primary_key=True),
                                  Column("resolver_id", Integer),
                                  Column("Key", Unicode(255), default=""),
                                  Column("Value", Unicode(2000), default=""),
                                  Column("Type", Unicode(2000), default=""),
                                  Column("Description", Unicode(2000), default=""),
                                  )

    resolverrealm_table = Table("resolverrealm", metadata,
                                Column("id", Integer,
                                       Sequence('resolverrealm_seq'),
                                       primary_key=True),
                                Column("resolver_id", Integer),
                                Column("realm_id", Integer),
                                Column("priority", Integer))

    #
    # LinOTP table definitions
    #

    linotp_token_table = Table('Token', metadata,
                               Column('LinOtpTokenId', Integer(),
                                      primary_key=True, nullable=False),
                               Column(
                                   'LinOtpTokenDesc', Unicode(80), default=''),
                               Column('LinOtpTokenSerialnumber', Unicode(
                                   40), default='', unique=True, nullable=False,
                                      index=True),
                               Column(
                                   'LinOtpTokenType', Unicode(30), default='HMAC',
                                   index=True),
                               Column(
                                   'LinOtpTokenInfo', Unicode(2000), default=''),
                               Column(
                                   'LinOtpTokenPinUser', Unicode(512), default=''),
                               Column(
                                   'LinOtpTokenPinUserIV', Unicode(32),
                                   default=''),
                               Column(
                                   'LinOtpTokenPinSO', Unicode(512), default=''),
                               Column(
                                   'LinOtpTokenPinSOIV', Unicode(32), default=''),
                               Column(
                                   'LinOtpIdResolver', Unicode(120), default='',
                                   index=True),
                               Column(
                                   'LinOtpIdResClass', Unicode(120), default=''),
                               Column(
                                   'LinOtpUserid', Unicode(320), default='',
                                   index=True),
                               Column(
                                   'LinOtpSeed', Unicode(32), default=''),
                               Column(
                                   'LinOtpOtpLen', Integer(), default=6),
                               Column(
                                   'LinOtpPinHash', Unicode(512), default=''),
                               Column(
                                   'LinOtpKeyEnc', Unicode(1024), default=''),
                               Column(
                                   'LinOtpKeyIV', Unicode(32), default=''),
                               Column(
                                   'LinOtpMaxFail', Integer(), default=10),
                               Column(
                                   'LinOtpIsactive', Boolean(), default=True),
                               Column(
                                   'LinOtpFailCount', Integer(), default=0),
                               Column('LinOtpCount', Integer(), default=0),
                               Column(
                                   'LinOtpCountWindow', Integer(), default=10),
                               Column(
                                   'LinOtpSyncWindow', Integer(), default=1000)
                               )

    linotp_engine = create_engine(config_obj.LINOTP_URI)
    privacyidea_engine = create_engine(config_obj.PRIVACYIDEA_URI)

    linotp_session = sessionmaker(bind=linotp_engine)()
    privacyidea_session = sessionmaker(bind=privacyidea_engine)()

    conn_linotp = linotp_engine.connect()
    conn_pi = privacyidea_engine.connect()

    def insert_chunks(conn, table, values, chunk_size=100000, record_name="records"):
        """
        Split **values** into chunks of size **chunk_size** and insert them sequentially.
        """
        values_length = len(values)
        for chunk in range(0, values_length, chunk_size):
            print('Insert {} {} to {} ...'.format(record_name, chunk,
                                                  min(chunk + chunk_size, values_length) - 1))
            try:
                conn.execute(table.insert(), values[chunk:chunk+chunk_size])
            except Exception as err:
                t = 'Failed to insert chunk: {0!s}'.format(err)
                warnings.append(t)
                print(t)

    # Values to be imported
    token_values = []
    tokeninfo_values = []
    tokenrealm_values = []
    tokenowner_values = []

    warnings = []

    # Process Assignments in table "tokenrealm"
    if config_obj.MIGRATE.get("assignments"):
        s = select([realm_table])
        result = conn_pi.execute(s)
        for r in result:
            realm_id_map[r["name"]] = r["id"]

        s = select([resolver_table])
        result = conn_pi.execute(s)
        for r in result:
            resolver_id_map[r["name"]] = r["id"]

        print("Realm-Map: {}".format(realm_id_map))
        print("Resolver-Map: {}".format(resolver_id_map))

    # Process Tokens

    if config_obj.MIGRATE.get("tokens"):
        s = select([linotp_token_table])
        result = conn_linotp.execute(s)

        i = 0
        for r in result:
            i = i + 1
            print("processing token #{1!s}: {0!s}".format(r["LinOtpTokenSerialnumber"], i))
            # Adapt type
            ttype = r["LinOtpTokenType"]
            if ttype.lower() == "hmac":
                ttype = "HOTP"
            # Adapt resolver
            linotp_resolver = r["LinOtpIdResClass"].split(".")[-1]

            # Adapt resolver type
            resolver_type = ""
            if r["LinOtpIdResClass"]:
                resolver_type = r['LinOtpIdResClass'].split(".")[1]
                resolver_type = resolver_map.get(resolver_type)

            # Adapt tokeninfo
            ti = {}
            if r["LinOtpTokenInfo"]:
                ti = json.loads(r["LinOtpTokenInfo"])

            if config_obj.MIGRATE.get("assignments"):
                user_pin = r['LinOtpTokenPinUser']
                user_pin_iv = r['LinOtpTokenPinUserIV']
                # Map the LinOTP-Resolver to the PI-Resolver
                resolver = config_obj.ASSIGNMENTS.get("resolver").get(linotp_resolver)
                if not resolver and linotp_resolver:
                    warnings.append("No mapping defined for the LinOTP "
                                    "resolver: {0!s}".format(linotp_resolver))
                resolver_type = resolver_type
                user_id = r['LinOtpUserid']
                if config_obj.ASSIGNMENTS.get("convert_endian"):
                    print(" +--- converting UUID {0!s}".format(user_id))
                    user_id = re.sub(UUID_MATCH_PATTERN, UUID_REPLACE_PATTERN, user_id)
                    print("  +-- to              {0!s}".format(user_id))
            else:
                user_pin = None
                user_pin_iv = None
                resolver = None
                resolver_type = None
                user_id = None

            key_enc = r['LinOtpKeyEnc']
            key_iv = r['LinOtpKeyIV']
            if config_obj.REENCRYPT:
                print(" +--- reencrypting token...")
                key_enc, key_iv = reencrypt(key_enc, key_iv,
                                            config_obj.LINOTP_ENC_KEY,
                                            config_obj.PRIVACYIDEA_ENC_KEY)

            token_values.append(dict(
                description=r["LinOtpTokenDesc"],
                serial=r["LinOtpTokenSerialnumber"],
                tokentype=ttype,
                user_pin=user_pin,
                user_pin_iv=user_pin_iv,
                so_pin=r['LinOtpTokenPinSO'],
                so_pin_iv=r['LinOtpTokenPinSOIV'],
                pin_seed=r['LinOtpSeed'],
                pin_hash=r['LinOtpPinHash'],
                key_enc=key_enc,
                key_iv=key_iv,
                maxfail=r['LinOtpMaxFail'],
                active=r['LinOtpIsactive'],
                failcount=r['LinOtpFailCount'],
                count=r['LinOtpCount'],
                count_window=r['LinOtpCountWindow'],
                sync_window=r['LinOtpSyncWindow'],
                # We add the user_id only that we can use it for the tokenowner!
                user_id=user_id,
                resolver=resolver
            ))

            if config_obj.MIGRATE.get("tokeninfo") and ti:
                # Add tokeninfo for this token
                for (k, v) in ti.items():
                    if k in config_obj.MIGRATE.get("tokeninfo_unicode_keys", []):
                        v = v.encode('ascii', 'ignore').decode('utf8')
                    tokeninfo_values.append(dict(
                        serial=r["LinOtpTokenSerialnumber"],
                        Key=k, Value=v,
                        token_id=r["LinOtpTokenId"]))
                    print(" +--- processing tokeninfo {0!s}".format(k))
                if config_obj.NEW_TOKENINFO:
                    print(" +--- processing new tokeninfo")
                    for k, v in config_obj.NEW_TOKENINFO.items():
                        tokeninfo_values.append(dict(
                            serial=r["LinOtpTokenSerialnumber"],
                            Key=k, Value=v,
                            token_id=r["LinOtpTokenId"]))

        print()
        print("Adding {} tokens...".format(len(token_values)))
        # Insert into database without the user_id
        insert_chunks(conn_pi, token_table,
                      [dict_without_keys(d, ["user_id", "resolver"]) for d in token_values],
                      config_obj.INSERT_CHUNK_SIZE, record_name="token records")

        # fetch the new token_id's in privacyIDEA and write them to the
        # token serial id map.
        s = select([token_table])
        result = conn_pi.execute(s)
        for r in result:
            token_serial_id_map[r["serial"]] = r["id"]

        # rewrite the id's in the token_values list
        for i in range(0, len(token_values)):
            token_values[i]["id"] = token_serial_id_map[token_values[i]["serial"]]

        if config_obj.MIGRATE.get("tokeninfo"):
            # Now we have to rewrite the token_id in the tokeninfo_values
            for ti in tokeninfo_values:
                ti["token_id"] = token_serial_id_map[ti["serial"]]
                del ti["serial"]

            print("Adding {} token infos...".format(len(tokeninfo_values)))
            insert_chunks(conn_pi, tokeninfo_table, tokeninfo_values,
                          config_obj.INSERT_CHUNK_SIZE, "tokeninfo records")

    if config_obj.MIGRATE.get("assignments"):
        # If the token is assigned, we also need to create an entry for tokenrealm
        # We need to determine the realm_id for this resolver!
        for token in token_values:
            token_id = token.get("id")
            resolver = token.get("resolver")
            user_id = token.get("user_id")
            if resolver and user_id:
                realm = config_obj.ASSIGNMENTS.get("realm").get(resolver)
                realm_id = realm_id_map.get(realm)
                print("Assigning token {} for resolver {} to realm_id {} "
                      "(realm {})".format(token_id, resolver, realm_id, realm))
                tokenrealm_values.append(dict(token_id=token_id,
                                              realm_id=realm_id))

                # Now we need to fill the tokenowner table, so that the tokens are actually
                # assigned to  user
                tokenowner_values.append(dict(token_id=token_id,
                                              realm_id=realm_id,
                                              resolver=resolver,
                                              user_id=user_id))
            else:
                # The token has no resolver and thus is not assigned
                for tokenrealm in config_obj.ASSIGNMENTS.get("unassigned_tokens", []):
                    realm_id = realm_id_map.get(tokenrealm)
                    if realm_id:
                        tokenrealm_values.append(dict(token_id=token_id,
                                                      realm_id=realm_id))

        print("Adding {} tokenrealms...".format(len(tokenrealm_values)))
        insert_chunks(conn_pi, tokenrealm_table, tokenrealm_values,
                      config_obj.INSERT_CHUNK_SIZE, record_name="tokenrealm records")

        print("Adding {} tokenowners...".format(len(tokenowner_values)))
        insert_chunks(conn_pi, tokenowner_table, tokenowner_values,
                      config_obj.INSERT_CHUNK_SIZE, record_name="tokenowner records")

    if warnings:
        print("We need to inform you about the following WARNINGS:")
        for warning in warnings:
            print(warning)


def usage():
    print("""
privacyidea-migrate-linotp.py --generate-example-config [--config <config file>]

    --generate-example-config, -g   Output an example config file. 
                                    This is a JSON file, that needs to be passed
                                    to this command.

    --config, -c <file>             The config file, that contains the complete
                                    configuration.

{0!s}
""".format(__doc__))


def read_enckey(encfile):
    with open(encfile, 'rb') as f:
        secret = f.read()[:32]
    return secret


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "gc:", ["generate-example-config", "config="])
    except getopt.GetoptError as e:
        print(str(e))
        sys.exit(1)

    config_file = None
    generate_config = False

    for o, a in opts:
        if o in ("-g", "--generate-example-config"):
            generate_config = True
            print(EXAMPLE_CONFIG_FILE)
        elif o in ("-c", "--config"):
            config_file = a
        else:
            print("Unknown parameter: {0!s}".format(o))
            sys.exit(3)

    if config_file:
        config_obj = Config(config_file)
        migrate(config_obj)
        sys.exit(0)

    else:
        if not generate_config:
            usage()
            sys.exit(1)


if __name__ == '__main__':
    main()



"""Add event handler conditions table

Revision ID: 3ae3c668f444
Revises: 5402fd96fbca
Create Date: 2016-07-20 12:18:55.643974

"""

# revision identifiers, used by Alembic.
revision = '3ae3c668f444'
down_revision = '5402fd96fbca'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError, InternalError


def upgrade():
    try:
        op.create_table('eventhandlercondition',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('eventhandler_id', sa.Integer(), nullable=True),
        sa.Column('Key', sa.Unicode(length=255), nullable=False),
        sa.Column('Value', sa.Unicode(length=2000), nullable=True),
        sa.Column('comparator', sa.Unicode(length=255), nullable=True),
        sa.ForeignKeyConstraint(['eventhandler_id'], ['eventhandler.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('eventhandler_id', 'Key', name='ehcix_1')
        )
    except (OperationalError, ProgrammingError, InternalError) as exx:
        if "duplicate column name" in str(exx.orig).lower():
            print("Good. Table eventhandlercondition already exists.")
        else:
            print("Table already exists")
            print(exx)

    except Exception as exx:
        print("Could not add Table eventhandlercondition")
        print (exx)


def downgrade():
    op.drop_table('eventhandlercondition')




# http://www.privacyidea.org
# (c) cornelius kölbel, privacyidea.org
#
# 2018-12-20 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Cleaning up the before_after method
# 2015-12-18 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Move before and after from api/token.py and system.py
#            to this central location
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This module contains the before and after routines for all
Flask endpoints.
It also contains the error handlers.
"""

from .lib.utils import (send_error, get_all_params, verify_auth_token)
from .container import container_blueprint
from ..lib.container import find_container_for_token, find_container_by_serial
from ..lib.framework import get_app_config_value
from ..lib.user import get_user_from_param
import logging
from .lib.utils import getParam
from flask import request, g
from privacyidea.lib.audit import getAudit
from flask import current_app
from privacyidea.lib.policy import PolicyClass
from privacyidea.lib.event import EventConfiguration
from privacyidea.lib.lifecycle import call_finalizers
from privacyidea.api.auth import (user_required, admin_required, jwtauth)
from privacyidea.lib.config import get_from_config, SYSCONF, ensure_no_config_object, get_privacyidea_node
from privacyidea.lib.token import get_token_type, get_token_owner
from privacyidea.api.ttype import ttype_blueprint
from privacyidea.api.validate import validate_blueprint
from .resolver import resolver_blueprint
from .policy import policy_blueprint
from .realm import realm_blueprint
from .realm import defaultrealm_blueprint
from .user import user_blueprint
from .audit import audit_blueprint
from .machineresolver import machineresolver_blueprint
from .machine import machine_blueprint
from .application import application_blueprint
from .caconnector import caconnector_blueprint
from .token import token_blueprint
from .system import system_blueprint
from .smtpserver import smtpserver_blueprint
from .radiusserver import radiusserver_blueprint
from .periodictask import periodictask_blueprint
from .privacyideaserver import privacyideaserver_blueprint
from .recover import recover_blueprint
from .register import register_blueprint
from .event import eventhandling_blueprint
from .smsgateway import smsgateway_blueprint
from .clienttype import client_blueprint
from .subscriptions import subscriptions_blueprint
from .monitoring import monitoring_blueprint
from .tokengroup import tokengroup_blueprint
from .serviceid import serviceid_blueprint
from .info import info_blueprint
from privacyidea.api.lib.postpolicy import postrequest, sign_response
from ..lib.error import (privacyIDEAError,
                         AuthError, UserError,
                         PolicyError, ResourceNotFoundError)
from privacyidea.lib.utils import get_client_ip, get_plugin_info_from_useragent
from privacyidea.lib.user import User
import datetime
import threading

log = logging.getLogger(__name__)


# ``before_app_request`` and ``teardown_app_request`` register the functions
# at the application, so it's sufficient to call them only for one blueprint.
# The decorated functions are called before and after *every* request.
@token_blueprint.before_app_request
def log_begin_request():
    log.debug("Begin handling of request {!r}".format(request.full_path))
    g.startdate = datetime.datetime.now()


@token_blueprint.teardown_app_request
def teardown_request(exc):
    try:
        if g.audit_object.has_data:
            g.audit_object.finalize_log()
    except AttributeError:
        # In certain error cases the before_request was not handled
        # completely so that we do not have an audit_object
        # Also during calling webui, there is no audit_object, yet.
        pass
    call_finalizers()
    log.debug("End handling of request {!r}".format(request.full_path))


@token_blueprint.before_request
@audit_blueprint.before_request
@system_blueprint.before_request
@info_blueprint.before_request
@user_required
def before_user_request():
    before_request()


@user_blueprint.before_request
@user_required
def before_userendpoint_request():
    before_request()
    # DEL /user/ has no realm parameter, and thus we need to create the user object this way.
    if not request.User and request.method == "DELETE":
        resolvername = getParam(request.all_data, "resolvername")
        username = getParam(request.all_data, "username")
        if resolvername and username:
            request.User = User(login=username, resolver=resolvername)


@resolver_blueprint.before_request
@machineresolver_blueprint.before_request
@machine_blueprint.before_request
@realm_blueprint.before_request
@defaultrealm_blueprint.before_request
@policy_blueprint.before_request
@application_blueprint.before_request
@smtpserver_blueprint.before_request
@eventhandling_blueprint.before_request
@periodictask_blueprint.before_request
@smsgateway_blueprint.before_request
@radiusserver_blueprint.before_request
@caconnector_blueprint.before_request
@privacyideaserver_blueprint.before_request
@client_blueprint.before_request
@subscriptions_blueprint.before_request
@monitoring_blueprint.before_request
@tokengroup_blueprint.before_request
@serviceid_blueprint.before_request
@admin_required
def before_admin_request():
    before_request()


@container_blueprint.before_request
def before_container_request():
    """
    This function is executed before a request to a container endpoint.
    It defines a list of endpoints that do not require an auth token and verifies the auth token for all other
    endpoints.
    The user is resolved either from the logged-in user or from the container owner. The user and some information
    about the container are stored in the audit log.
    Additionally, the policy object, the audit object and the event configuration object are set to the global variable.
    """
    # auth free endpoints
    ensure_no_config_object()
    request.all_data = get_all_params(request)
    auth_token_free_endpoints = ["/container/register/finalize",
                                 "/container/register/terminate/client",
                                 "/container/challenge",
                                 "/container/synchronize",
                                 "/container/rollover", ]
    is_auth_free = request.path in auth_token_free_endpoints

    # Get auth token
    auth_token = request.headers.get('PI-Authorization')
    if not auth_token:
        auth_token = request.headers.get('Authorization')

    # Verify auth token for all non-auth-token-free endpoints anyway
    if not is_auth_free:
        r = verify_auth_token(auth_token, ["user", "admin"])
        g.logged_in_user = {"username": r.get("username"),
                            "realm": r.get("realm"),
                            "role": r.get("role")}
        resolve_logged_in_user()
    else:
        if auth_token:
            # if an auth token is present, but it is not mandatory for the endpoint, write it in the params to be
            # verified later
            request.all_data["auth_token"] = auth_token
        # Client requests does not provide a logged-in user, hence it cannot be resolved is set to None
        request.User = None

    get_before_request_config()
    g.serial = None
    privacyidea_server = get_app_config_value("PI_AUDIT_SERVERNAME", get_privacyidea_node(request.host))
    # Get info about the container and resolve user from the container if not yet set
    container_serial = getParam(request.all_data, "container_serial")
    container = None
    container_type = None
    if container_serial and "*" not in container_serial:
        try:
            container = find_container_by_serial(container_serial)
        except ResourceNotFoundError:
            # The container serial might not exist
            pass
    if container:
        container_type = container.type
        if request.User is None or request.User.is_empty():
            owners = container.get_users()
            if len(owners) == 1:
                request.User = owners[0]

    if request.User:
        audit_username = request.User.login
        audit_realm = request.User.realm
        audit_resolver = request.User.resolver
    else:
        audit_realm = getParam(request.all_data, "realm")
        audit_resolver = getParam(request.all_data, "resolver")
        audit_username = getParam(request.all_data, "user")

    ua_name, ua_version, _ua_comment = get_plugin_info_from_useragent(request.user_agent.string)
    g.audit_object.log({"success": False,
                        "user": audit_username,
                        "realm": audit_realm,
                        "resolver": audit_resolver,
                        "container_serial": container_serial,
                        "container_type": container_type,
                        "client": g.client_ip,
                        "user_agent": ua_name,
                        "user_agent_version": ua_version,
                        "privacyidea_server": privacyidea_server,
                        "action": f"{request.method} {request.url_rule}",
                        "action_detail": "",
                        "thread_id": threading.current_thread().ident,
                        "info": ""})


def resolve_logged_in_user():
    """
    Resolve the user attributes (username, realm and resolver) from the logged-in user and set them in the request
    object.
    """
    if g.logged_in_user.get("role") == "user":
        # A user is calling this API. First thing we do is restricting the user parameter.
        # ...to restrict token view, audit view or token actions.
        request.all_data["user"] = g.logged_in_user.get("username")
        request.all_data["realm"] = g.logged_in_user.get("realm")

    try:
        request.User = get_user_from_param(request.all_data)
        # overwrite or set the resolver parameter in case of a logged-in user
        if g.logged_in_user.get("role") == "user":
            request.all_data["resolver"] = request.User.resolver
    except AttributeError:
        # Some endpoints do not need users OR e.g. the setPolicy endpoint
        # takes a list as the user object
        request.User = None
    except UserError:
        # In cases like the policy API, the parameter "user" is part of the
        # policy and will not resolve to a user object
        request.User = User()


def get_before_request_config():
    """
    Gets the policy object, the audit object and the event configuration object and sets them to the global flask
    variable. Additionally, reads the client IP and the HTTP headers from the request object and writes them to the
    global flask variable.
    """
    g.policy_object = PolicyClass()
    g.audit_object = getAudit(current_app.config, g.startdate)
    g.event_config = EventConfiguration()
    # access_route contains the ip addresses of all clients, hops and proxies.
    g.client_ip = get_client_ip(request, get_from_config(SYSCONF.OVERRIDECLIENT))
    # Save the HTTP header in the localproxy object
    g.request_headers = request.headers


def before_request():
    """
    This is executed before the request.

    user_required checks if there is a logged in admin or user

    The checks for ONLY admin are performed in api/system.py
    """
    # remove session from param and gather all parameters, either
    # from the Form data or from JSON in the request body.
    ensure_no_config_object()
    request.all_data = get_all_params(request)
    resolve_logged_in_user()
    get_before_request_config()

    privacyidea_server = get_app_config_value("PI_AUDIT_SERVERNAME", get_privacyidea_node(request.host))
    # Already get some typical parameters to log
    serial = getParam(request.all_data, "serial")
    if serial and "*" not in serial and "," not in serial:
        g.serial = serial
        tokentype = get_token_type(serial)
        if not request.User:
            # We determine the user object by the given serial number
            try:
                request.User = get_token_owner(serial) or User()
            except ResourceNotFoundError:
                # The serial might not exist! This would raise an exception
                pass

    else:
        g.serial = None
        tokentype = None

    # Container info
    container_serial = getParam(request.all_data, "container_serial")
    container = None
    container_type = None
    if container_serial and "*" not in container_serial:
        try:
            container = find_container_by_serial(container_serial)
        except ResourceNotFoundError:
            # The container serial might not exist
            pass
    elif serial and "*" not in serial:
        # Get container serial from token
        # if a serial list is given, check if all tokens are in the same container
        serial_list = serial.replace(" ", "").split(",")
        for idx, t_serial in enumerate(serial_list):
            try:
                container = find_container_for_token(t_serial)
            except ResourceNotFoundError:
                # The container serial might not exist
                continue
            if container:
                if idx == 0:
                    container_serial = container.serial
            if not container or container_serial != container.serial:
                # If at least one token is in a different or no container, we do not set the container attributes.
                container_serial = None
                break
    if container:
        container_type = container.type

    if request.User:
        audit_username = request.User.login
        audit_realm = request.User.realm
        audit_resolver = request.User.resolver
    else:
        audit_realm = getParam(request.all_data, "realm")
        audit_resolver = getParam(request.all_data, "resolver")
        audit_username = getParam(request.all_data, "user")

    ua_name, ua_version, _ua_comment = get_plugin_info_from_useragent(request.user_agent.string)
    g.audit_object.log({"success": False,
                        "serial": serial,
                        "user": audit_username,
                        "realm": audit_realm,
                        "resolver": audit_resolver,
                        "token_type": tokentype,
                        "container_serial": container_serial,
                        "container_type": container_type,
                        "client": g.client_ip,
                        "user_agent": ua_name,
                        "user_agent_version": ua_version,
                        "privacyidea_server": privacyidea_server,
                        "action": "{0!s} {1!s}".format(request.method, request.url_rule),
                        "action_detail": "",
                        "thread_id": "{0!s}".format(threading.current_thread().ident),
                        "info": ""})

    if g.logged_in_user.get("role") == "admin":
        # An administrator is calling this API
        g.audit_object.log({"administrator": g.logged_in_user.get("username")})
        # TODO: Check is there are realm specific admin policies, so that the
        # admin is only allowed to act on certain realms
        # If now realm is specified, we need to add "filterrealms".
        # If the admin tries to view realms, he is not allowed to, we need to
        #  raise an exception.


@system_blueprint.after_request
@resolver_blueprint.after_request
@realm_blueprint.after_request
@defaultrealm_blueprint.after_request
@policy_blueprint.after_request
@user_blueprint.after_request
@token_blueprint.after_request
@audit_blueprint.after_request
@application_blueprint.after_request
@machine_blueprint.after_request
@machineresolver_blueprint.after_request
@caconnector_blueprint.after_request
@smtpserver_blueprint.after_request
@eventhandling_blueprint.after_request
@radiusserver_blueprint.after_request
@smsgateway_blueprint.after_request
@periodictask_blueprint.after_request
@privacyideaserver_blueprint.after_request
@client_blueprint.after_request
@subscriptions_blueprint.after_request
@monitoring_blueprint.after_request
@ttype_blueprint.after_request
@validate_blueprint.after_request
@register_blueprint.after_request
@recover_blueprint.after_request
@tokengroup_blueprint.after_request
@serviceid_blueprint.after_request
@container_blueprint.after_request
@info_blueprint.after_request
@jwtauth.after_request
@postrequest(sign_response, request=request)
def after_request(response):
    """
    This function is called after a request
    :return: The response
    """
    # No caching!
    response.headers['Cache-Control'] = 'no-cache'
    return response


@system_blueprint.errorhandler(AuthError)
@realm_blueprint.app_errorhandler(AuthError)
@defaultrealm_blueprint.app_errorhandler(AuthError)
@resolver_blueprint.app_errorhandler(AuthError)
@policy_blueprint.app_errorhandler(AuthError)
@user_blueprint.app_errorhandler(AuthError)
@token_blueprint.app_errorhandler(AuthError)
@audit_blueprint.app_errorhandler(AuthError)
@application_blueprint.app_errorhandler(AuthError)
@smtpserver_blueprint.app_errorhandler(AuthError)
@eventhandling_blueprint.app_errorhandler(AuthError)
@subscriptions_blueprint.app_errorhandler(AuthError)
@monitoring_blueprint.app_errorhandler(AuthError)
@tokengroup_blueprint.app_errorhandler(AuthError)
@serviceid_blueprint.app_errorhandler(AuthError)
@container_blueprint.app_errorhandler(AuthError)
@info_blueprint.app_errorhandler(AuthError)
def auth_error(error):
    if "audit_object" in g:
        message = ''

        if hasattr(error, 'message'):
            message = error.message

        if hasattr(error, 'details'):
            if error.details:
                if 'message' in error.details:
                    message = '{}|{}'.format(message, error.details['message'])

        g.audit_object.add_to_log({"info": message}, add_with_comma=True)
    return send_error(error.message,
                      error_code=error.id,
                      details=error.details), 401


@system_blueprint.errorhandler(PolicyError)
@realm_blueprint.app_errorhandler(PolicyError)
@defaultrealm_blueprint.app_errorhandler(PolicyError)
@resolver_blueprint.app_errorhandler(PolicyError)
@policy_blueprint.app_errorhandler(PolicyError)
@user_blueprint.app_errorhandler(PolicyError)
@token_blueprint.app_errorhandler(PolicyError)
@audit_blueprint.app_errorhandler(PolicyError)
@application_blueprint.app_errorhandler(PolicyError)
@smtpserver_blueprint.app_errorhandler(PolicyError)
@eventhandling_blueprint.app_errorhandler(PolicyError)
@register_blueprint.app_errorhandler(PolicyError)
@recover_blueprint.app_errorhandler(PolicyError)
@subscriptions_blueprint.app_errorhandler(PolicyError)
@monitoring_blueprint.app_errorhandler(PolicyError)
@ttype_blueprint.app_errorhandler(PolicyError)
@tokengroup_blueprint.app_errorhandler(PolicyError)
@serviceid_blueprint.app_errorhandler(PolicyError)
@container_blueprint.app_errorhandler(PolicyError)
@info_blueprint.app_errorhandler(PolicyError)
def policy_error(error):
    if "audit_object" in g:
        g.audit_object.add_to_log({"info": error.message}, add_with_comma=True)
    return send_error(error.message, error_code=error.id), 403


@system_blueprint.app_errorhandler(ResourceNotFoundError)
@realm_blueprint.app_errorhandler(ResourceNotFoundError)
@defaultrealm_blueprint.app_errorhandler(ResourceNotFoundError)
@resolver_blueprint.app_errorhandler(ResourceNotFoundError)
@policy_blueprint.app_errorhandler(ResourceNotFoundError)
@user_blueprint.app_errorhandler(ResourceNotFoundError)
@token_blueprint.app_errorhandler(ResourceNotFoundError)
@audit_blueprint.app_errorhandler(ResourceNotFoundError)
@application_blueprint.app_errorhandler(ResourceNotFoundError)
@smtpserver_blueprint.app_errorhandler(ResourceNotFoundError)
@eventhandling_blueprint.app_errorhandler(ResourceNotFoundError)
@register_blueprint.app_errorhandler(ResourceNotFoundError)
@recover_blueprint.app_errorhandler(ResourceNotFoundError)
@subscriptions_blueprint.app_errorhandler(ResourceNotFoundError)
@ttype_blueprint.app_errorhandler(ResourceNotFoundError)
@tokengroup_blueprint.errorhandler(ResourceNotFoundError)
@serviceid_blueprint.errorhandler(ResourceNotFoundError)
@container_blueprint.app_errorhandler(ResourceNotFoundError)
@info_blueprint.app_errorhandler(ResourceNotFoundError)
def resource_not_found_error(error):
    """
    This function is called when an ResourceNotFoundError occurs.
    It sends a 404.
    """
    if "audit_object" in g:
        g.audit_object.log({"info": error.message})
    return send_error(error.message, error_code=error.id), 404


@system_blueprint.app_errorhandler(privacyIDEAError)
@realm_blueprint.app_errorhandler(privacyIDEAError)
@defaultrealm_blueprint.app_errorhandler(privacyIDEAError)
@resolver_blueprint.app_errorhandler(privacyIDEAError)
@policy_blueprint.app_errorhandler(privacyIDEAError)
@user_blueprint.app_errorhandler(privacyIDEAError)
@token_blueprint.app_errorhandler(privacyIDEAError)
@audit_blueprint.app_errorhandler(privacyIDEAError)
@application_blueprint.app_errorhandler(privacyIDEAError)
@smtpserver_blueprint.app_errorhandler(privacyIDEAError)
@eventhandling_blueprint.app_errorhandler(privacyIDEAError)
@register_blueprint.app_errorhandler(privacyIDEAError)
@recover_blueprint.app_errorhandler(privacyIDEAError)
@subscriptions_blueprint.app_errorhandler(privacyIDEAError)
@monitoring_blueprint.app_errorhandler(privacyIDEAError)
@ttype_blueprint.app_errorhandler(privacyIDEAError)
@tokengroup_blueprint.app_errorhandler(privacyIDEAError)
@serviceid_blueprint.app_errorhandler(privacyIDEAError)
@container_blueprint.app_errorhandler(privacyIDEAError)
@info_blueprint.app_errorhandler(privacyIDEAError)
def privacyidea_error(error):
    """
    This function is called when an privacyIDEAError occurs.
    These are not that critical exceptions.
    """
    if "audit_object" in g:
        g.audit_object.log({"info": str(error)})
    return send_error(str(error), error_code=error.id), 400


@system_blueprint.app_errorhandler(NotImplementedError)
@realm_blueprint.app_errorhandler(NotImplementedError)
@defaultrealm_blueprint.app_errorhandler(NotImplementedError)
@resolver_blueprint.app_errorhandler(NotImplementedError)
@policy_blueprint.app_errorhandler(NotImplementedError)
@user_blueprint.app_errorhandler(NotImplementedError)
@token_blueprint.app_errorhandler(NotImplementedError)
@audit_blueprint.app_errorhandler(NotImplementedError)
@application_blueprint.app_errorhandler(NotImplementedError)
@smtpserver_blueprint.app_errorhandler(NotImplementedError)
@eventhandling_blueprint.app_errorhandler(NotImplementedError)
@register_blueprint.app_errorhandler(NotImplementedError)
@recover_blueprint.app_errorhandler(NotImplementedError)
@subscriptions_blueprint.app_errorhandler(NotImplementedError)
@monitoring_blueprint.app_errorhandler(NotImplementedError)
@ttype_blueprint.app_errorhandler(NotImplementedError)
@tokengroup_blueprint.app_errorhandler(NotImplementedError)
@serviceid_blueprint.app_errorhandler(NotImplementedError)
@container_blueprint.app_errorhandler(NotImplementedError)
def not_implemented_error(error):
    """
    This function is called when an NotImplementedError occurs.
    """
    if "audit_object" in g:
        g.audit_object.log({"info": str(error)})
    return send_error(str(error), error_code=-501), 501


# other errors
@system_blueprint.app_errorhandler(500)
@realm_blueprint.app_errorhandler(500)
@defaultrealm_blueprint.app_errorhandler(500)
@resolver_blueprint.app_errorhandler(500)
@policy_blueprint.app_errorhandler(500)
@user_blueprint.app_errorhandler(500)
@token_blueprint.app_errorhandler(500)
@audit_blueprint.app_errorhandler(500)
@application_blueprint.app_errorhandler(500)
@smtpserver_blueprint.app_errorhandler(500)
@eventhandling_blueprint.app_errorhandler(500)
@register_blueprint.app_errorhandler(500)
@recover_blueprint.app_errorhandler(500)
@subscriptions_blueprint.app_errorhandler(500)
@monitoring_blueprint.app_errorhandler(500)
@ttype_blueprint.app_errorhandler(500)
@tokengroup_blueprint.app_errorhandler(500)
@serviceid_blueprint.app_errorhandler(500)
@container_blueprint.app_errorhandler(500)
@info_blueprint.app_errorhandler(500)
def internal_error(error):
    """
    This function is called when an internal error (500) occurs.
    i.e. if a normal exception, that is not inherited from privacyIDEAError
    occurs.
    """
    if "audit_object" in g:
        g.audit_object.log({"info": str(error)})
    return send_error(str(error), error_code=-500), 500



"""v3.8: Add thread_id to audit_log

Revision ID: 006d4747f858
Revises: d3c0f0403a84
Create Date: 2022-11-30 22:37:42.163199

"""

# revision identifiers, used by Alembic.
revision = '006d4747f858'
down_revision = 'd3c0f0403a84'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError, InternalError


def upgrade():
    try:
        op.add_column('pidea_audit', sa.Column('thread_id', sa.Unicode(length=20), nullable=True))
    except (OperationalError, ProgrammingError, InternalError) as exx:
        print("Looks like the thread_id already exists in the pidea_audit table.")
        print(exx)
    except Exception as exx:
        print("Could not add thread_id to pidea_audit table.")
        print (exx)


def downgrade():
    op.drop_column('pidea_audit', 'thread_id')



from setuptools import setup, find_packages
import os
import stat
import sys

VERSION = "3.11.2"

# Taken from kennethreitz/requests/setup.py
package_directory = os.path.realpath(os.path.dirname(__file__))


def get_file_contents(file_path):
    """Get the context of the file using full path name."""
    content = ""
    try:
        full_path = os.path.join(package_directory, file_path)
        content = open(full_path, 'r').read()
    except:
        print("### could not open file {0!r}".format(file_path), file=sys.stderr)
    return content


def get_file_list(file_path):
    full_path = os.path.join(package_directory, file_path)
    file_list = os.listdir(full_path)
    # now we need to add the path to the files
    return [file_path + f for f in file_list]


install_requires = [
    "argon2_cffi",
    "beautifulsoup4[lxml]",
    "cbor2",
    "configobj",
    "croniter",
    "cryptography",
    "defusedxml",
    "Flask",
    "Flask-Babel",
    "Flask-Migrate",
    "Flask-SQLAlchemy",
    "Flask-Versioned",
    "feedparser",
    "google-auth",
    "grpcio",
    "huey[redis]",
    "ldap3<2.9",
    "MarkupSafe",
    "netaddr",
    "passlib[bcrypt]",
    "protobuf",
    "pydash",
    "PyJWT",
    "PyMySQL",
    "pyOpenSSL<=24.0.0",
    "pyrad",
    "python-dateutil",
    "python-gnupg",
    "PyYAML",
    "requests",
    "segno",
    "smpplib",
    "SQLAlchemy<2.0",
    "webauthn"
]


def get_man_pages(dir):
    """
    Get man pages in a directory.
    :param dir:
    :return: list of file names
    """
    files = os.listdir(dir)
    r_files = []
    for file in files:
        if file.endswith(".1"):
            r_files.append(dir + "/" + file)
    return r_files


def get_scripts(dir):
    """
    Get files that are executable
    :param dir:
    :return: list of file names
    """
    files = os.listdir(dir)
    r_files = []
    for file in files:
        if os.stat(dir + "/" + file)[stat.ST_MODE] & stat.S_IEXEC:
            r_files.append(dir + "/" + file)
    return r_files


setup(
    name='privacyIDEA',
    version=VERSION,
    description='privacyIDEA: multifactor authentication management system',
    author='privacyidea.org',
    license='AGPLv3',
    author_email='cornelius@privacyidea.org',
    url='https://www.privacyidea.org',
    keywords='OTP, two factor authentication, management, security',
    python_requires='>=3.8',
    packages=find_packages(),
    scripts=get_scripts("tools"),
    entry_points={
        'console_scripts': [
            'privacyidea-token-janitor = privacyidea.cli.privacyideatokenjanitor:cli',
            'pi-manage = privacyidea.cli.pimanage:cli',
            'privacyidea-standalone = privacyidea.cli.tools.standalone:cli',
            'privacyidea-get-serial = privacyidea.cli.tools.get_serial:byotp_call',
            'privacyidea-usercache-cleanup = privacyidea.cli.tools.usercache_cleanup:delete_call',
            'privacyidea-get-unused-tokens = privacyidea.cli.tools.get_unused_tokens:cli',
            'privacyidea-expired-users = privacyidea.cli.tools.expired_users:expire_call',
            'privacyidea-cron = privacyidea.cli.tools.cron:cli',
            'pi-tokenjanitor = privacyidea.cli.pitokenjanitor:cli'
        ]},
    extras_require={
        'doc': ["Pallets-Sphinx-Themes",
                "Sphinx",
                "sphinxcontrib-httpdomain",
                "sphinxcontrib-spelling"],
        'test': ["mock",
                 "pyparsing",
                 "pytest",
                 "pytest-cov",
                 "responses",
                 "testfixtures"],
        'postgres': ['psycopg2'],
        'hsm': ['PyKCS11'],
        'kerberos': ['gssapi']
    },
    install_requires=install_requires,
    include_package_data=True,
    data_files=[('etc/privacyidea/',
                 ['deploy/apache/privacyideaapp.wsgi',
                  'deploy/privacyidea/dictionary']),
                ('share/man/man1', get_man_pages("tools")),
                ('lib/privacyidea/migrations',
                 ["migrations/alembic.ini",
                  "migrations/env.py",
                  "migrations/README",
                  "migrations/script.py.mako"]),
                ('lib/privacyidea/migrations/versions',
                 get_file_list("migrations/versions/")),
                ('lib/privacyidea/', ['requirements.txt'])
                ],
    classifiers=["Framework :: Flask",
                 "License :: OSI Approved :: "
                 "GNU Affero General Public License v3",
                 "Programming Language :: Python",
                 "Development Status :: 5 - Production/Stable",
                 "Topic :: Internet",
                 "Topic :: Security",
                 "Topic :: System ::"
                 " Systems Administration :: Authentication/Directory",
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: 3.9',
                 'Programming Language :: Python :: 3.10',
                 'Programming Language :: Python :: 3.11',
                 'Programming Language :: Python :: 3.12'
                 ],
    zip_safe=False,
    long_description=get_file_contents('README.rst')
)



#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#  HMAC-OTP (RFC 4226)
#  Copyright (C) LSE Leading Security Experts GmbH, Weiterstadt
#  Written by Max Vozeler <max.vozeler@lsexperts.de>
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
"""
  Description:  HOTP basic functions
"""
  
import hmac
import logging
import struct
import binascii

from hashlib import sha1

from privacyidea.lib.utils import hexlify_and_unicode
from privacyidea.lib.log import log_with
from privacyidea.lib.crypto import safe_compare


log = logging.getLogger(__name__)


class HmacOtp(object):

    def __init__(self, secObj=None, counter=0, digits=6, hashfunc=sha1):
        self.secretObj = secObj
        self.counter = int(counter)
        self.digits = digits
        self.hashfunc = hashfunc

    def hmac(self,
             counter=None,
             key=None,
             challenge=None):
        """

        :param counter:
        :param key:
        :param challenge: The datainput for OCRA
        :type challenge: hex string
        :return:
        :rtype: bytes
        """
        # log.error("hmacSecret()")
        if counter is None:
            counter = self.counter

        # When using a counter, we can only use 64bit as data_input.
        # When we allow a raw data_input, we could use 160bit or more.
        if not challenge:
            data_input = struct.pack(">Q", counter)
        else:
            data_input = binascii.unhexlify(challenge)

        if key is None:
            dig = self.secretObj.hmac_digest(data_input, self.hashfunc)
        else:
            dig = hmac.new(key, data_input, self.hashfunc).digest()

        return dig

    def truncate(self, digest):
        offset = digest[-1] & 0x0f

        binary = (digest[offset + 0] & 0x7f) << 24
        binary |= (digest[offset + 1] & 0xff) << 16
        binary |= (digest[offset + 2] & 0xff) << 8
        binary |= (digest[offset + 3] & 0xff)

        return binary % (10 ** self.digits)

    def generate(self,
                 counter=None,
                 inc_counter=True,
                 key=None,
                 do_truncation=True,
                 challenge=None):
        """

        :param counter:
        :param inc_counter:
        :param key:
        :param do_truncation:
        :param challenge: hexlified challenge
        :return:
        :rtype: str
        """
        if counter is None:
            counter = self.counter

        if challenge:
            hmac = self.hmac(challenge=challenge, key=key)
        else:
            hmac = self.hmac(counter=counter, key=key)
        if do_truncation:
            otp = str(self.truncate(hmac))
            """  fill in the leading zeros  """
            sotp = (self.digits - len(otp)) * "0" + otp
        else:
            sotp = hexlify_and_unicode(hmac)
            
        if inc_counter:
            self.counter = counter + 1
        return sotp

    @log_with(log)
    def checkOtp(self, anOtpVal, window, symetric=False):
        """

        :param anOtpVal: The OTP value to check
        :type anOtpVal: str
        :param window:
        :param symetric:
        :return: -1 if the OTP doesn't match or the current counter
        :rtype: int
        """
        res = -1
        start = self.counter
        end = self.counter + window
        if symetric is True:
            # changed window/2 to window for TOTP
            start = self.counter - (window)
            start = 0 if (start < 0) else start
            end = self.counter + (window)

        log.debug("OTP range counter: {0!r} - {1!r}".format(start, end))
        for c in range(start, end):
            otpval = self.generate(c)
            #log.debug("calculating counter {0!r}".format(c))

            if safe_compare(otpval, anOtpVal):
                res = c
                break
        # return -1 or the counter
        return res



#  privacyIDEA
#  Jul 18, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__="""The SSHKeyTokenClass provides a TokenClass that stores the public
SSH key. This can be used to manage SSH keys and retrieve the public ssh key
to import it to authorized keys files.

The code is tested in tests/test_lib_tokens_ssh
"""

import logging
from privacyidea.lib import _
from privacyidea.api.lib.utils import getParam
from privacyidea.lib.error import TokenAdminError
from privacyidea.lib.log import log_with
from privacyidea.lib.tokenclass import TokenClass, ROLLOUTSTATE, AUTHENTICATIONMODE
from privacyidea.lib.policy import SCOPE, ACTION, GROUP

log = logging.getLogger(__name__)


optional = True
required = False


##TODO: We should save a fingerprint of the SSH Key in the encrypted OTP
# field, so that we can be sure, that the public ssh key was not changed in
# the database!


class SSHkeyTokenClass(TokenClass):
    """
    The SSHKeyTokenClass provides a TokenClass that stores the public
    SSH key. This can be used to manage SSH keys and retrieve the public ssh key
    to import it to authorized keys files.
    """
    mode = [AUTHENTICATIONMODE.AUTHENTICATE]
    using_pin = False

    def __init__(self, db_token):
        TokenClass.__init__(self, db_token)
        self.set_type("sshkey")

    @staticmethod
    def get_class_type():
        return "sshkey"

    @staticmethod
    def get_class_prefix():
        return "SSHK"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        returns a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dictionary
        """
        res = {'type': 'sshkey',
               'title': 'SSHkey Token',
               'description': _('SSH Public Key: The public SSH key.'),
               'config': {},
               'user': ['enroll'],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin", "user"],
               'policy': {
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of SSH keys assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _(
                               "The user may only have this maximum number of active SSH keys assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               },
               }
        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res

        return ret

    def update(self, param):
        """
        The key holds the public ssh key and this is required
        
        The key probably is of the form "ssh-rsa BASE64 comment"
        """
        # We need to save the token, so that we can later add the tokeninfo
        # Otherwise we might not have created the DB entry, yet and we would
        # be missing the token.id
        self.token.save()

        getParam(param, "sshkey", required)

        key_elem = param.get("sshkey").split(" ", 2)
        if key_elem[0] not in ["ssh-rsa", "ssh-ed25519", "ecdsa-sha2-nistp256",
                               "sk-ecdsa-sha2-nistp256@openssh.com", "sk-ssh-ed25519@openssh.com"]:
            self.token.rollout_state = ROLLOUTSTATE.BROKEN
            self.token.save()
            raise TokenAdminError("The keytype you specified is not supported.")

        if len(key_elem) < 2:
            self.token.rollout_state = ROLLOUTSTATE.BROKEN
            self.token.save()
            raise TokenAdminError("Missing key.")

        key_type = key_elem[0]
        key = key_elem[1]
        if len(key_elem) > 2:
            key_comment = key_elem[2]
        else:
            key_comment = ""
        
        # convert key to hex
        self.add_tokeninfo("ssh_key", key, value_type="password")
        self.add_tokeninfo("ssh_type", key_type)
        self.add_tokeninfo("ssh_comment", key_comment)

        # call the parents function
        TokenClass.update(self, param)
        
    @log_with(log)
    def get_sshkey(self):
        """
        returns the public SSH key
        
        :return: SSH pub key
        :rtype: string
        """
        ti = self.get_tokeninfo()
        key_type = ti.get("ssh_type")
        key_comment = ti.get("ssh_comment")
        # get the ssh key directly, otherwise it will not be decrypted
        sshkey = self.get_tokeninfo("ssh_key")
        r = "{0!s} {1!s}".format(key_type, sshkey)
        if key_comment:
            r += " " + key_comment
        return r



__doc__ = """This is the event handler module for custom user attributes.
You can set or delete custom your attributes

"""

from privacyidea.lib.eventhandler.base import BaseEventHandler
from privacyidea.lib import _
import logging

from privacyidea.lib.user import User

log = logging.getLogger(__name__)


class ACTION_TYPE(object):
    """
    Allowed actions
    """
    SET_CUSTOM_USER_ATTRIBUTES = "set_custom_user_attributes"
    DELETE_CUSTOM_USER_ATTRIBUTES = "delete_custom_user_attributes"


class USER_TYPE(object):
    """
    Allowed user types
    """
    TOKENOWNER = "tokenowner"
    LOGGED_IN_USER = "logged_in_user"


class CustomUserAttributesHandler(BaseEventHandler):
    """
    The CustomUserAttributesHandler is an EventHandler which can set/change/delete custom User-Attributes.
    """

    identifier = "CustomUserAttributes"
    description = "This eventhandler can set and delete custom_user_attributes"

    @property
    def allowed_positions(cls):
        """
        This returns the allowed positions of the event handler definition.

        :return: list of allowed positions
        """
        return ["post", "pre"]

    @property
    def actions(cls):
        """
        This method returns a dictionary of allowed actions and possible
        options in this handler module.

        :return: dict with actions
        """
        actions = {
            ACTION_TYPE.SET_CUSTOM_USER_ATTRIBUTES: {
                "user": {
                    'type': 'str',
                    'required': True,
                    'description': _('The user for whom the custom attribute should be set.'),
                    "value": [
                        USER_TYPE.TOKENOWNER,
                        USER_TYPE.LOGGED_IN_USER,
                    ]},
                "attrkey": {
                    'type': 'str',
                    'description': _('The key of the custom user attribute that should be set.')},
                "attrvalue": {
                    'type': 'str',
                    'description': _('The value of the custom user attribute.')}
            },
            ACTION_TYPE.DELETE_CUSTOM_USER_ATTRIBUTES: {
                "user": {
                    'type': 'str',
                    'required': True,
                    'description': _('The user from which the custom attribute should be deleted.'),
                    "value": [
                        USER_TYPE.TOKENOWNER,
                        USER_TYPE.LOGGED_IN_USER,
                    ]},
                "attrkey": {
                    'type': 'str',
                    'description': _('The key of the custom user attribute that should be deleted.')}
        }}
        return actions

    def do(self, action, options=None):
        """
        This method executes the defined action in the given event.

        :param action: The action to perform
        :type action: str
        :param options: Contains the flask parameters g,  attrkey and attrvalue
        :type options: dict
        :return:
        """
        g = options.get("g")
        request = options.get("request")
        handler_def = options.get("handler_def")
        handler_options = handler_def.get("options", {})
        user_type = handler_options.get("user", USER_TYPE.TOKENOWNER)
        tokenowner = self._get_tokenowner(request)
        if user_type == USER_TYPE.TOKENOWNER and not tokenowner.is_empty():
            user = tokenowner
        elif user_type == USER_TYPE.LOGGED_IN_USER and hasattr(g, 'logged_in_user'):
            user = User(login=g.logged_in_user.get('username'),
                        realm=g.logged_in_user.get('realm'))
        else:
            log.warning("Unable to determine the user for handling the custom "
                        "attribute! action: {0!s}, handler: {1!s}".format(action, handler_def))
            return False

        attrkey = handler_options.get("attrkey")
        attrvalue = handler_options.get("attrvalue")
        if action.lower() == "set_custom_user_attributes":
            ret = user.set_attribute(attrkey, attrvalue)
        elif action.lower() == "delete_custom_user_attributes":
            ret = user.delete_attribute(attrkey)
        else:
            log.warning('Unknown action value: {0!s}'.format(action))
            ret = False

        return ret



#  2015-07-09 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Fix mOTP reuse of OTP values
#  2015-01-27 Rewrite due to flask migration
#             Cornelius Kölbel <cornelius@privacyidea.org>
#  2014-10-03 Add getInitDetail
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
#  May 08, 2014 Cornelius Kölbel
#  Sept 16, 2014 Cornelius Kölbel, added key generation for Token2
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#
#  privacyIDEA is a fork of LinOTP
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This code implements the motp one time password algorithm
described in motp.sourceforge.net.

The code is tested in tests/test_lib_tokens_motp
"""
from .mOTP import mTimeOtp
from privacyidea.lib.apps import create_motp_url
from privacyidea.lib.tokenclass import TokenClass
from privacyidea.lib.log import log_with
from privacyidea.lib.utils import create_img, is_true
from privacyidea.api.lib.utils import getParam
from privacyidea.lib.crypto import generate_otpkey
from privacyidea.lib.decorators import check_token_locked
import traceback
import logging
from privacyidea.lib import _
from privacyidea.lib.policy import SCOPE, ACTION, GROUP

optional = True
required = False
log = logging.getLogger(__name__)


class MotpTokenClass(TokenClass):

    @staticmethod
    def get_class_type():
        return "motp"

    @staticmethod
    def get_class_prefix():
        return "PIMO"

    @staticmethod
    def get_class_info(key=None, ret='all'):
        """
        returns a subtree of the token definition
        Is used by lib.token.get_token_info

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype : dict or string
        """

        res = {'type': 'motp',
               'title': 'mOTP Token',
               'description': _('mOTP: Classical mobile One Time Passwords.'),
               'init': {'page': {'html': 'motptoken.mako',
                                 'scope': 'enroll', },
                        'title': {'html': 'motptoken.mako',
                                  'scope': 'enroll.title'},
                        },
               'config': {'page': {'html': 'motptoken.mako',
                                   'scope': 'config'},
                          'title': {'html': 'motptoken.mako',
                                    'scope': 'config.title', },
                          },
               'user': ['enroll'],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin", "user"],
               'policy': {
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of mOTP tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of active mOTP tokens assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               }
               }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res

        return ret

    @log_with(log)
    def __init__(self, db_token):
        """
        constructor - create a token object

        :param db_token: instance of the orm db object
        :type db_token:  orm object
        """
        TokenClass.__init__(self, db_token)
        self.set_type("motp")
        self.hKeyRequired = True
        return

    @log_with(log)
    def get_init_detail(self, params=None, user=None):
        """
        to complete the token normalisation, the response of the initialization
        should be built by the token specific method, the getInitDetails
        """
        response_detail = TokenClass.get_init_detail(self, params, user)
        otpkey = self.init_details.get('otpkey')
        if otpkey:
            if user is not None:
                try:
                    motp_url = create_motp_url(otpkey,
                                               user.login, user.realm,
                                               serial=self.get_serial())
                    response_detail["motpurl"] = {"description": _("URL for mOTP "
                                                                   "token"),
                                                  "value": motp_url,
                                                  "img": create_img(motp_url)
                                                  }
                except Exception as ex:   # pragma: no cover
                    log.debug("{0!s}".format(traceback.format_exc()))
                    log.error('failed to set motp url: {0!r}'.format(ex))

        return response_detail

    @log_with(log)
    def update(self, param, reset_failcount=True):
        """
        update - process initialization parameters

        :param param: dict of initialization parameters
        :type param: dict

        :return: nothing
        """
        if self.hKeyRequired is True:
            genkey = is_true(getParam(param, "genkey", optional))
            if not param.get('keysize'):
                param['keysize'] = 16
            if genkey:
                otpKey = generate_otpkey(param['keysize'])
                del param['genkey']
            else:
                # genkey not set: check otpkey is given
                # this will raise an exception if otpkey is not present
                otpKey = getParam(param, "otpkey", required)

            param['otpkey'] = otpKey

        # motp token specific
        mOTPPin = getParam(param, "motppin", required)
        self.token.set_user_pin(mOTPPin)

        TokenClass.update(self, param, reset_failcount)

        return

    @log_with(log)
    @check_token_locked
    def check_otp(self, anOtpVal, counter=None, window=None, options=None):
        """
        validate the token otp against a given otpvalue

        :param anOtpVal: the to be verified otpvalue
        :type anOtpVal:  str
        :param counter: the counter state, that should be verified
        :type counter: int
        :param window: the counter +window, which should be checked
        :type window: int
        :param options: the dict, which could contain token specific info
        :type options: dict
        :return: the counter state or -1
        :rtype: int
        """
        otplen = self.token.otplen

        # otime contains the previous verification time
        # the new one must be newer than this!
        oCount = self.get_otp_count()
        secretHOtp = self.token.get_otpkey()
        window = self.token.count_window
        secretPin = self.token.get_user_pin()

        log.debug("original counter %s", oCount)

        mtimeOtp = mTimeOtp(secretHOtp, secretPin, oCount, otplen)
        res = mtimeOtp.checkOtp(anOtpVal, window, options=options)

        if res != -1 and oCount != 0 and res <= oCount:
            log.warning("a previous OTP value was used again! former "
                        "tokencounter: %i, presented counter %i" %
                        (oCount, res))
            res = -1
            return res

        if res != -1:
            # on success, we have to save the last attempt
            self.set_otp_count(res)

        return res



"""v3.8: Merging revisions

Revision ID: d3c0f0403a84
Revises: ('fabcf24d9304', 'a28f2733897b')
Create Date: 2022-11-30 22:37:19.986083

"""

# revision identifiers, used by Alembic.
revision = 'd3c0f0403a84'
down_revision = ('fabcf24d9304', 'a28f2733897b')

from alembic import op
import sqlalchemy as sa


def upgrade():
    pass


def downgrade():
    pass
